# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch
from torch import Tensor
import json
import os
from typing import List, Union
import logging

from .dtypes import *
from .triton_kernels import *
from .triton_kernels.utils import gpu_supports_float16_acc
from .triton_kernels import utils

from .bitpack import (
    pack_weights_over_cols_triton,
    pack_weights_over_cols_torch,
)

from .quant_utils import (
    scale_activations_per_token,
    scale_activations_mxfp8,
    scale_activations_mxfp4,
    scale_activations_nvfp4,
)

import threading

FILE_LOCK = threading.Lock()

logger = logging.getLogger(__name__)

###################################################################################################################################
# Global variables
###################################################################################################################################
GEMLITE_ACC_DTYPE = {
    DType.FP16: DType.FP16 if gpu_supports_float16_acc() else DType.FP32,
    DType.BF16: DType.FP32,
    DType.FP32: DType.FP32,
    DType.FP8: DType.FP32,
    DType.FP8e4: DType.FP32,
    DType.FP8e4nuz: DType.FP32,
    DType.FP8e5: DType.FP32,
    DType.FP8e5nuz: DType.FP32,
    DType.INT8: DType.INT32,
    DType.MXFP16: DType.FP32,
    DType.MXBF16: DType.FP32,
    DType.MXFP8: DType.FP32,
    DType.MXFP4: DType.FP32,
    DType.NVFP4: DType.FP32,
}

GEMLITE_TRITON_KERNELS = [
    gemv,
    gemv_revsplitK,
    gemv_splitK,
    gemm_splitK,  # gemm_splitK / gemm_splitK_persistent
    gemm,
]

GEMLITE_TRITON_MAPPING = {
    kernel.matmul_type: kernel for kernel in GEMLITE_TRITON_KERNELS
}
GEMLITE_MATMUL_TYPES = [kernel.matmul_type for kernel in GEMLITE_TRITON_KERNELS]
GEMLITE_MATMUL_TYPES_MAPPING = {
    GEMLITE_MATMUL_TYPES[i]: i for i in range(len(GEMLITE_MATMUL_TYPES))
}
GEMLITE_TRITON_CONFIG_CACHE = {}  # Global config cache for all the kernels
_GROUP_SIZE_WARNED = False

###################################################################################
# Utils


# Main function to cache kernel autotune config
def cache_kernel_config(kernel, num_keys):
    kernel_cache = kernel.cache
    k_config = {}
    if len(kernel_cache) > 0:
        for k in kernel_cache:
            key = list(k[:num_keys])
            key[0] = utils.get_closest_m(key[0])  # restrict batch-size
            key = tuple(key)
            k_config[str(key)] = kernel_cache[k].all_kwargs()
    return k_config


# Set M autotune logic
def set_autotune_setting(fct):  # fct = lambda M: M for max-autotune
    utils.get_closest_m = fct


# set default packing format
def set_packing_bitwidth(packing_bitwidth: int):
    GemLiteLinearTriton.PACKING_BITWIDTH = packing_bitwidth


# Set accumulation dtype
def set_acc_dtype(dtype):
    global GEMLITE_ACC_DTYPE
    assert dtype in [DType.FP16, DType.FP32], (
        "Invalid dtype (should be DType.FP16 or DType.FP32)."
    )
    GEMLITE_ACC_DTYPE[DType.FP16] = dtype


# Return the default gemv kernel to use for M==1
def get_default_gemv(W_nbits: int, mx_dtype: bool = False) -> str:
    # TODO: adapt mx for IS_HIP = True
    if mx_dtype:
        return "GEMM_SPLITK"  # TODO: fix mxf bugs in GEMV outputs garbage.
    else:
        return "GEMV_REVSPLITK" if (W_nbits < 8) else "GEMV_SPLITK"


# matmul type selection logic
def get_matmul_type(batch_size: int, W_nbits: int, mx_dtype: bool = False):
    if batch_size > 64:
        return "GEMM"
    if batch_size > 1:
        return "GEMM_SPLITK"
    else:
        return get_default_gemv(W_nbits, mx_dtype)


#######################################################################################################################
def enable_activation_scaling(batch_size):
    """
    This functions enables adaptive activation quantization based on the batch-size.
    For example, it would only do dynamic activation in more compute-bound settings (batch_size >=32 / 64).
    Only works with the MXFP format - use with A8W4_MXFP/A4W4_MXFP.
    """
    return True
    # return batch_size >= 32


# Main functional forward call
@torch.library.custom_op("gemlite::forward_functional", mutates_args=())
def forward_functional(
    x: Tensor,
    bias: Union[None, Tensor],
    tensor_args: List[Tensor],
    meta_args: List[int],
    matmul_type: int = -1,  # -1: auto, >=0: manual
) -> Tensor:

    data_contiguous = bool(meta_args[1])
    W_nbits = meta_args[1]
    out_features = tensor_args[0].shape[1]

    # Get type_id for autotune: use same autotune for float16/bfloat16
    input_dtype = meta_args[5]
    input_dtype = DType.FP16.value if (input_dtype == DType.BF16.value) else input_dtype
    input_dtype = (
        DType.MXFP16.value if (input_dtype == DType.MXBF16.value) else input_dtype
    )
    type_id = input_dtype * 100 + W_nbits

    if not x.is_contiguous():
        x = x.contiguous()

    batch_size = x.numel() // x.shape[-1]
    orig_shape = x.shape
    out_shape = x.shape[:-1] + (out_features,)
    x_dtype = x.dtype

    scaled_activations = bool(meta_args[0]) and enable_activation_scaling(batch_size)
    # Dynamic activation quantization
    scales_x = None
    if scaled_activations:
        input_dtype = DType(meta_args[5])
        channel_scale_mode = meta_args[9]

        if input_dtype in FP8_INT8_DTYPES:  # INT8 / FP8
            x, scales_x = scale_activations_per_token(
                x, w_dtype=DTYPE_TO_TORCH[input_dtype.value]
            )

        elif input_dtype in [DType.MXFP8] and channel_scale_mode == 4:  # MXPF8
            x, scales_x = scale_activations_mxfp8(x, w_dtype=torch.float8_e4m3fn)

        elif (
            input_dtype in [DType.MXFP8] and channel_scale_mode == 2
        ):  # MXPF8 with post-scale for the activations
            x, scales_x = scale_activations_per_token(x, w_dtype=torch.float8_e4m3fn)

        elif input_dtype in [DType.MXFP4] and channel_scale_mode == 4:  # MXPF4
            x, scales_x = scale_activations_mxfp4(x)

        elif input_dtype in [DType.NVFP4] and channel_scale_mode == 4:  # NVPF4: TODO
            x, scales_x = scale_activations_nvfp4(x)

    x = x.view(-1, x.shape[-1])

    if matmul_type >= 0:
        matmul_type_str = GEMLITE_MATMUL_TYPES[matmul_type]
    else:
        matmul_type_str = get_matmul_type(
            x.shape[0], W_nbits, is_mx_dtype(input_dtype)
        )  # batch_size, W_nbits, is_mx_dtype

    out = (
        GEMLITE_TRITON_MAPPING[matmul_type_str]
        .forward(x, *tensor_args, scales_x, *meta_args[1:-1], data_contiguous, type_id)
        .view(out_shape)
    )

    if bias is not None:
        out += bias

    return out


@torch.library.register_fake("gemlite::forward_functional")
def forward_functional_fake(
    x: Tensor,
    bias: Union[None, Tensor],
    tensor_args: List[Tensor],
    meta_args: List[int],
    matmul_type: int = -1,
) -> Tensor:
    out_features = tensor_args[0].shape[1]
    return torch.empty(x.shape[:-1] + (out_features,), device=x.device, dtype=x.dtype)


#######################################################################################################################
# Main class
class GemLiteLinearTriton(torch.nn.Module):
    SUPPORTED_BITS_TRITON = [1, 2, 4, 8, 16]
    SUPPORTED_DTYPES = [
        DType.FP16,
        DType.BF16,
        DType.FP32,
        DType.FP8,
        DType.FP8e4,
        DType.FP8e4nuz,
        DType.FP8e5,
        DType.FP8e5nuz,
        DType.INT8,
        DType.MXFP16,
        DType.MXBF16,
        DType.MXFP8,
        DType.MXFP4,
        DType.NVFP4,
    ]
    MIN_SIZE = 32
    PACKING_BITWIDTH = 32  # Default packing bitwidth

    def __init__(
        self,
        W_nbits=4,
        group_size=64,
        in_features=None,
        out_features=None,
        input_dtype=DType.FP16,
        output_dtype=DType.FP16,
        acc_dtype=None,
        scaled_activations=False,
    ):
        global _GROUP_SIZE_WARNED

        super().__init__()

        if W_nbits not in GemLiteLinearTriton.SUPPORTED_BITS_TRITON:
            raise NotImplementedError(
                "Only "
                + str(GemLiteLinearTriton.SUPPORTED_BITS_TRITON)
                + " W_nbits are supported."
            )

        if in_features is not None and out_features is not None:
            if (in_features % GemLiteLinearTriton.MIN_SIZE != 0) or (
                in_features % group_size != 0 if (group_size is not None) else False
            ):
                raise NotImplementedError(
                    "Invalid input shapes: "
                    + str(in_features)
                    + " , "
                    + str(out_features)
                    + ". in_features should be divisible by 32 or the group_size"
                )

        # Warning: Input dtype should be the same as dequantize() weights dtype.
        if input_dtype not in GemLiteLinearTriton.SUPPORTED_DTYPES:
            raise NotImplementedError("Unsupport input dtype: " + str(input_dtype))

        if group_size is not None:
            if group_size < 16:
                raise NotImplementedError("Only group_size >= 16 is supported.")

        group_size = 1 if (group_size is None) else group_size

        self.in_features = in_features
        self.out_features = out_features
        self.orig_shape = (out_features, in_features)
        self.W_nbits = W_nbits
        self.group_size = group_size
        self.unpack_mask = 2**self.W_nbits - 1
        self.elements_per_sample = None
        self.signature = (in_features, out_features, W_nbits, group_size)

        self.input_dtype = input_dtype
        self.output_dtype = output_dtype
        self.compute_dtype = DTYPE_TO_TORCH[self.input_dtype.value]
        self.meta_dtype = input_dtype

        # Accumulation
        self.acc_dtype = (
            GEMLITE_ACC_DTYPE[self.input_dtype] if (acc_dtype is None) else acc_dtype
        )

        # Scales activations
        if self.compute_dtype in [torch.float16, torch.bfloat16, torch.float32]:
            self.scaled_activations = False
        else:
            self.scaled_activations = scaled_activations

        # Default forward
        self.forward = self.forward_auto_no_warmup

    def load_state_dict(self, state_dict, strict=True, assign=False):
        self.W_q = state_dict.pop("W_q", None)
        self.bias = state_dict.pop("bias", None)
        self.scales = state_dict.pop("scales", None)
        self.zeros = state_dict.pop("zeros", None)
        self.metadata = state_dict.pop("metadata", None)
        self.orig_shape = state_dict.pop("orig_shape", None)

        self.metadata = [v.item() for v in self.metadata]
        self.orig_shape = (v.item() for v in self.orig_shape)

        (
            self.scaled_activations,
            self.W_nbits,
            self.group_size,
            self.unpack_mask,
            self.elements_per_sample,
            self.input_dtype,
            self.output_dtype,
            self.acc_dtype,
            self.meta_dtype,
            self.channel_scale_mode,
            self.W_group_mode,
            self.data_contiguous,
        ) = self.metadata

        self.input_dtype = DType(self.input_dtype)
        self.output_dtype = DType(self.output_dtype)
        self.acc_dtype = DType(self.acc_dtype)
        self.meta_dtype = DType(self.meta_dtype)

        self.out_features, self.in_features = self.orig_shape
        self.compute_dtype = DTYPE_TO_TORCH[self.input_dtype.value]
        self.scaled_activations = bool(self.scaled_activations)
        self.data_contiguous = bool(self.data_contiguous)

    # Make sure to feed UINT8 W_q for packing
    def pack(
        self,
        W_q: Tensor,
        scales: Tensor,
        zeros: Union[Tensor, int],
        bias: Union[Tensor, None] = None,
        fma_mode: bool = True,
        contiguous: Union[int, None] = None,
        packing_bitwidth: Union[int, None] = None,
    ):

        # Check inputs
        raise_unsupported = False
        if zeros is not None and self.input_dtype == DType.INT8:
            if isinstance(zeros, Tensor):
                if zeros.mean() != zeros.int().float().mean():
                    raise_unsupported = True
            elif isinstance(zeros, float):
                raise_unsupported = True

        if raise_unsupported:
            raise Exception("INT8 inputs is not compatible with floating-point zeros.")

        # Set packing bitwidth
        if packing_bitwidth is None:
            packing_bitwidth = GemLiteLinearTriton.PACKING_BITWIDTH

        # Only 8-bit packing is supported for microscaling
        if is_mx_dtype(self.input_dtype):
            packing_bitwidth = 8

        # Non-packed weights
        self.W_q = None
        if W_q.dtype in [torch.int8] or W_q.is_floating_point():
            if W_q.dtype in [torch.float32]:
                assert self.W_nbits == 32, "Invalid fp32 weights."
            elif W_q.dtype in [torch.float16, torch.bfloat16]:
                assert self.W_nbits == 16, "Invalid fp16 weights."
            else:
                assert self.W_nbits == 8, "Invalid 8-bit weights."

            self.W_q = W_q.t()
            self.elements_per_sample = 1

            if contiguous is None:
                contiguous = False

        # Packed weigths
        if W_q.dtype == torch.uint8:
            _pack_weights_over_cols = (
                pack_weights_over_cols_triton
                if (W_q.device.type == "cuda")
                else pack_weights_over_cols_torch
            )

            self.W_q, self.elements_per_sample = _pack_weights_over_cols(
                W_q.view(self.orig_shape),
                W_nbits=self.W_nbits,
                packing_bitwidth=packing_bitwidth,
                transpose=True,
            )  # Over-K

            if contiguous is None:
                if is_mx_dtype(self.input_dtype):
                    contiguous = False
                else:
                    contiguous = True

        if self.W_q is None:
            raise Exception("Weights were not packed, please check your W_q.dtype")

        # Bias / device
        self.device = self.W_q.device
        self.bias = None if (bias is None) else bias.to(device=self.device)

        # initial values
        self.W_group_mode = -1
        self.channel_scale_mode = 0

        # FP16 x FP16 / FP8 x FP8 / INT8 x INT8 - no meta-data case
        if (scales is None) and (zeros is None):
            self.zeros = None
            self.scales = None
            self.W_group_mode = 0
            self.channel_scale_mode = 2 if self.scaled_activations else 0

        # The rest of the use-cases require some kind of meta-data
        if scales is not None:
            self.scales = scales.view((self.out_features, -1)).t()
        else:
            self.scales = None

        # channel-wise scaling
        self.meta_is_channelwise = (
            False if (self.scales is None) else self.scales.numel() == self.out_features
        )

        # Symmetric no shift
        if zeros is None:
            self.zeros = None
            self.W_group_mode = 2 if (self.scales is not None) else 0
        else:
            # Asymmetric or Symmetric with shift
            if isinstance(zeros, torch.Tensor):
                if fma_mode and (
                    self.meta_is_channelwise is False
                ):  # W ~ Wq * scales + zeros
                    self.zeros = (
                        (-zeros.float() * scales.float())
                        .to(zeros.dtype)
                        .view((self.out_features, -1))
                        .t()
                    )
                    self.W_group_mode = 4
                else:  # W ~ (Wq - zeros) * scales
                    self.zeros = zeros.view((self.out_features, -1)).t()
                    self.W_group_mode = 3
            else:  # Integer
                self.zeros = int(zeros)
                if self.scales is not None:
                    self.W_group_mode = 3  # Symmetric with shift
                else:
                    self.W_group_mode = 1  # Shift only with integer

        assert self.W_group_mode > -1, "Invalid scales/zeros settings."

        # weight-only
        if (self.scaled_activations == False) and (self.meta_is_channelwise == True):
            self.channel_scale_mode = 1
            self.W_group_mode = (
                1 if (self.zeros is not None) else 0
            )  # only with fma_mode=False

        # activation-only
        if (self.scaled_activations == True) and (self.meta_is_channelwise == False):
            self.channel_scale_mode = 2

        # weight + activation mode
        if (self.scaled_activations == True) and (self.meta_is_channelwise == True):
            self.channel_scale_mode = 3
            self.W_group_mode = (
                1 if (self.zeros is not None) else 0
            )  # only with fma_mode=False

        if self.channel_scale_mode in [1, 3]:
            assert self.W_group_mode not in [3, 4], (
                "Can't use channel_scale_mode with W_group_mode == 3 or 4."
            )

        # if(self.input_dtype == DType.INT8):
        #     assert self.W_group_mode in [1], "Only channel-wise symmetric quantization is supported for INT8 inputs."

        # Dummy values
        if isinstance(self.zeros, int):  # Union[Tensor, int] not supported by custom op
            self.zeros = torch.tensor(self.zeros, dtype=torch.int32, device=self.device)
        if self.zeros is None:
            self.zeros = torch.tensor([[]], dtype=torch.int32, device=self.device)
        if self.scales is None:
            self.scales = torch.tensor([[]], dtype=torch.int32, device=self.device)

        # Force contiguous
        if contiguous:
            self.data_contiguous = True
            self.W_q = self.W_q.contiguous()
        else:
            self.data_contiguous = False

        if isinstance(self.scales, torch.Tensor):
            self.scales = self.scales.contiguous()
        if isinstance(self.zeros, torch.Tensor):
            self.zeros = self.zeros.contiguous()

        # MX dtypes scaling
        if self.input_dtype in [DType.MXFP16, DType.MXBF16, DType.MXFP8, DType.MXFP4]:
            self.scales = self.scales.to(torch.float8_e8m0fnu).view(torch.uint8)
        if self.input_dtype in [DType.NVFP4]:
            self.scales = self.scales.to(torch.float8_e4m3fn)
        if is_mx_dtype(self.input_dtype):
            self.scales = self.scales.T
            self.W_group_mode = 2
            self.channel_scale_mode = 0

        if self.scales is not None:
            self.meta_dtype = TORCH_TO_DTYPE[self.scales.dtype]

        # Register tensors as buffers
        self.W_q = torch.nn.Parameter(self.W_q, requires_grad=False)
        self.bias = (
            torch.nn.Parameter(self.bias, requires_grad=False)
            if self.bias is not None
            else None
        )
        self.scales = torch.nn.Parameter(self.scales, requires_grad=False)
        self.zeros = torch.nn.Parameter(self.zeros, requires_grad=False)

        # Register metadata
        self.metadata = torch.nn.Parameter(
            torch.tensor(self.get_meta_args(), device=self.device, dtype=torch.int32),
            requires_grad=False,
        )

        self.orig_shape = torch.nn.Parameter(
            torch.tensor(
                [self.out_features, self.in_features],
                device=self.device,
                dtype=torch.int32,
            ),
            requires_grad=False,
        )

        return self

    # Return the main arguments
    def get_tensor_args(self):
        return [self.W_q, self.scales, self.zeros]

    def get_meta_args(self):
        return [
            int(self.scaled_activations),
            self.W_nbits,
            self.group_size,
            self.unpack_mask,
            self.elements_per_sample,
            self.input_dtype.value,
            self.output_dtype.value,
            self.acc_dtype.value,
            self.meta_dtype.value,
            self.channel_scale_mode,
            self.W_group_mode,
            int(self.data_contiguous),
        ]

    # #Main manual call
    def forward_manual(self, x: Tensor, matmul_type: str = "GEMM") -> Tensor:
        return forward_functional(
            x,
            self.bias,
            self.get_tensor_args(),
            self.get_meta_args(),
            GEMLITE_MATMUL_TYPES_MAPPING[matmul_type],
        )

    # #Main auto call without exhaustive search
    def forward_auto_no_warmup(self, x: Tensor) -> Tensor:
        return forward_functional(
            x,
            self.bias,
            self.get_tensor_args(),
            self.get_meta_args(),
        )

    @staticmethod
    def cache_config(filename: str):
        global GEMLITE_TRITON_CONFIG_CACHE
        # Load existing cache if available
        try:
            with FILE_LOCK, open(filename, "r") as json_file:
                config = json.load(json_file)
        except:
            config = {}

        # Can't use GEMLITE_TRITON_MAPPING for some reason kernel.cache is empty
        _GEMLITE_TRITON_MAPPING = {}
        from .triton_kernels.gemv_kernels import gemv

        _GEMLITE_TRITON_MAPPING["GEMV"] = gemv

        from .triton_kernels.gemv_revsplitK_kernels import gemv_revsplitK

        _GEMLITE_TRITON_MAPPING["GEMV_REVSPLITK"] = gemv_revsplitK

        from .triton_kernels.gemv_splitK_kernels import gemv_splitK

        _GEMLITE_TRITON_MAPPING["GEMV_SPLITK"] = gemv_splitK

        from .triton_kernels.gemm_splitK_kernels import gemm_splitK

        _GEMLITE_TRITON_MAPPING["GEMM_SPLITK"] = gemm_splitK

        from .triton_kernels.gemm_kernels import gemm

        _GEMLITE_TRITON_MAPPING["GEMM"] = gemm

        for name in _GEMLITE_TRITON_MAPPING:
            if name not in config:
                config[name] = {}

            if name in GEMLITE_TRITON_CONFIG_CACHE:
                config[name].update(GEMLITE_TRITON_CONFIG_CACHE[name])

            for kernel in _GEMLITE_TRITON_MAPPING[name].kernel:
                config[name].update(
                    cache_kernel_config(kernel, 6)
                )  # 5: len(prune_keys)

        # Save combined cache
        with FILE_LOCK, open(filename, "w") as json_file:
            json.dump(config, json_file)

    @staticmethod
    def load_config(filename: str, print_error: bool = True, overwrite: bool = False):
        global GEMLITE_TRITON_CONFIG_CACHE
        if filename is None:
            return False
        try:
            with FILE_LOCK, open(filename, "r") as json_file:
                config = json.load(json_file)
                if overwrite:
                    GEMLITE_TRITON_CONFIG_CACHE = config
                else:
                    for name in config:
                        if name not in GEMLITE_TRITON_CONFIG_CACHE:
                            GEMLITE_TRITON_CONFIG_CACHE[name] = {}
                        GEMLITE_TRITON_CONFIG_CACHE[name].update(config[name])

        except Exception as e:
            if print_error:
                logger.error(f"Failed to load the cache file '{filename}': {e}")
            return False
        return True

    @staticmethod
    def reset_config():
        global GEMLITE_TRITON_CONFIG_CACHE
        GEMLITE_TRITON_CONFIG_CACHE = {}


###################################################################################################################################
###################################################################################################################################
GemLiteLinear = GemLiteLinearTriton  # Triton by default


# Setting default config
def get_default_cache_config():
    root_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "configs/")

    def get_tags(path):
        return [f.split(".")[0] for f in os.listdir(path)]

    name = torch.cuda.get_device_properties(0).name.lower().replace(" ", "_")
    tags = get_tags(root_path)
    tags.sort(key=len, reverse=True)

    selected_tag = None
    for tag in tags:
        if tag in name:
            selected_tag = os.path.join(root_path, tag + ".json")
            break

    return selected_tag


selected_tag = get_default_cache_config()
if GemLiteLinear.load_config(selected_tag):
    logger.warning("Loaded " + selected_tag + " config.")



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch
import triton
import math
import triton.language as tl
from triton.runtime import driver
from ..dtypes import *


@triton.jit
def swizzle_tile_v1(
    pid,
    M,
    N,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
):
    grid_m = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n = tl.cdiv(N, BLOCK_SIZE_N)
    width = GROUP_SIZE_M * grid_n
    group_id = pid // width
    group_size = tl.minimum(grid_m - group_id * GROUP_SIZE_M, GROUP_SIZE_M)
    pid_m = group_id * GROUP_SIZE_M + (pid % group_size)
    pid_n = (pid % width) // group_size
    return pid_m, pid_n


@triton.jit
def swizzle_tile_v2(
    pid,
    M,
    N,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
):
    grid_m = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n = tl.cdiv(N, BLOCK_SIZE_N)
    width = GROUP_SIZE_M * grid_m
    group_id = pid // width
    group_size = tl.minimum(grid_n - group_id * GROUP_SIZE_M, GROUP_SIZE_M)
    pid_n = group_id * GROUP_SIZE_M + (pid % group_size)
    pid_m = (pid % width) // group_size
    return pid_m, pid_n


@triton.jit
def swizzle_tile_v3(
    pid,
    M,
    N,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
):
    pid_m = pid % tl.cdiv(M, BLOCK_SIZE_M)
    pid_n = pid // tl.cdiv(M, BLOCK_SIZE_M)
    grid_m = tl.cdiv(M, BLOCK_SIZE_M)
    grid_n = tl.cdiv(N, BLOCK_SIZE_N)
    return tl.swizzle2d(pid_m, pid_n, grid_m, grid_n, GROUP_SIZE_M)


@triton.jit
def swizzle_tile_persistent(
    tile_id, num_pid_in_group, num_pid_m, GROUP_SIZE_M: tl.constexpr
):
    group_id = tile_id // num_pid_in_group
    first_pid_m = group_id * GROUP_SIZE_M
    group_size_m = min(num_pid_m - first_pid_m, GROUP_SIZE_M)
    pid_m = first_pid_m + (tile_id % group_size_m)
    pid_n = (tile_id % num_pid_in_group) // group_size_m
    return pid_m, pid_n


swizzle_tile = swizzle_tile_v1


@triton.jit
def linear_tile(
    pid,
    M,
    N,
    BLOCK_SIZE_M: tl.constexpr,
    BLOCK_SIZE_N: tl.constexpr,
    GROUP_SIZE_M: tl.constexpr,
):
    pid_m = pid % tl.cdiv(M, BLOCK_SIZE_M)
    pid_n = pid // tl.cdiv(M, BLOCK_SIZE_M)
    return pid_m, pid_n


#################################################################################################################
@triton.jit
def dequantize(
    b,
    scales,
    zeros,
    q_shift,
    meta_dtype,
    unpack_mask,
    elements_per_sample: tl.constexpr,
    W_group_mode: tl.constexpr,
    zero_is_scalar: tl.constexpr,
):
    # Unpack
    if elements_per_sample > 1:
        b = (b >> q_shift) & unpack_mask  # int32 -> int32

    if W_group_mode == 1:  # Shift
        b = b.to(meta_dtype) - zeros

    if W_group_mode == 2:
        b = b.to(meta_dtype) * scales  # Symmetric no shift (Grouped)

    if (
        W_group_mode == 3
    ):  # Asymmetric / Symmetric with shift(Grouped - (b - zeros) * scales)
        # b = (b - zeros) * scales
        if zero_is_scalar:
            b = (b - zeros).to(meta_dtype) * scales
        else:
            b = (b.to(meta_dtype) - zeros) * scales

    if W_group_mode == 4:
        b = tl.fma(
            b.to(meta_dtype), scales, zeros
        )  # Asymmetric (Grouped - b*scales + zeros)

    return b


@triton.jit
def atomic_add_cas(ptr, value, Lock, mask=None, sem: tl.constexpr = "release"):
    while tl.atomic_cas(Lock, 0, 1, sem=sem) == 1:
        pass
    tl.store(ptr, tl.load(ptr, mask=mask) + value, mask=mask)
    tl.debug_barrier()
    tl.atomic_xchg(Lock, 0)


def init_to_zero(name):
    return lambda nargs: nargs[name].zero_()


def next_power_of_2(v):
    return 2 ** int(math.ceil(math.log2(v)))


def is_divisible(dividend, divisor):
    return dividend % divisor == 0


def is_hip():
    return triton.runtime.driver.active.get_current_target().backend == "hip"


def gpu_has_more_shared_memory(
    ref_gpus=["a100", "h100", "h200", "h20", "h800", "b100", "b200"],
):
    gpu_name = torch.cuda.get_device_properties(0).name.lower()
    return True in [g in gpu_name for g in ref_gpus]


def gpu_supports_float16_acc(
    ref_gpus=[
        "5090",
        "5080",
        "5070",
        "5060",
        "4090",
        "4080",
        "4070",
        "4060",
        "3090",
        "3080",
        "3070",
        "3060",
        "2080",
        "2070",
    ],
):
    gpu_name = torch.cuda.get_device_properties(0).name.lower()
    return True in [g in gpu_name for g in ref_gpus]


def gpu_supports_bfloat16_atomicadd():
    # Triton tl.atomic_add doens't support bfloat16 even for Hopper and above.
    # return torch.cuda.get_device_capability()[0] >= 9 #Hopper and above
    return False


NUM_SMS = torch.cuda.get_device_properties("cuda").multi_processor_count


def get_num_SMs(device):
    # return torch.cuda.get_device_properties(device).multi_processor_count
    return NUM_SMS  # cache it to avoid driver ping - should be ok as multi-gpu systems tend to have the same devices


# Only powers of 2
def generate_autotune_lookup_v1(max_m=16384):
    return [
        min(2 ** int(math.ceil(math.log2(M))), max_m) if (M > 0) else 0
        for M in range(max_m + 1)
    ]


# Powers of 2 but also (power of 2 + next power of 2) / divisor,
def generate_autotune_lookup_v2(
    max_m=16384, min_split=32, divisors=[2, 4], mode="next", include_vllm_config=False
):
    lookup = [0] * (max_m + 1)
    autotune_vals = set()

    i = 0
    while (val := 2**i) <= max_m:
        autotune_vals.add(val)
        next_val = 2 ** (i + 1)
        if val >= min_split and next_val <= max_m:
            for d in divisors:
                interpolated = (val + next_val) // d
                autotune_vals.add(interpolated)
        i += 1

    if include_vllm_config:
        autotune_vals.update(
            [
                1,
                2,
                4,
                8,
                16,
                24,
                32,
                40,
                48,
                56,
                64,
                72,
                80,
                88,
                96,
                104,
                112,
                120,
                128,
                136,
                144,
                152,
                160,
                168,
                176,
                184,
                192,
                200,
                208,
                216,
                224,
                232,
                240,
                248,
                256,
                384,
                512,
            ]
        )

    sorted_vals = sorted(autotune_vals)

    for m in range(max_m + 1):
        if mode == "next":
            lookup[m] = min(
                (x for x in sorted_vals if x >= m), default=None
            )  # Next-value
        elif mode == "closest":
            lookup[m] = min(
                sorted_vals, key=lambda x: (abs(x - m), x < m)
            )  # Closest-Value
        else:
            raise Exception("Invalid mode.")
    return lookup


M_MAXVAL = 4096  # 1024, 4096, 16384
M_MAPPING = generate_autotune_lookup_v2(M_MAXVAL, mode="next")


def get_closest_m(M):
    return M_MAPPING[M] if M <= M_MAXVAL else M_MAXVAL


def get_gpu_shared_memory():
    return driver.active.utils.get_device_properties(0).get("max_shared_mem", 0)


###################################################################################
# Cached results to avoid runtime driver calls

IS_HIP = is_hip()
NATIVE_ATOMIC = gpu_supports_bfloat16_atomicadd()



from functools import lru_cache
from typing import Optional

import torch
import torch.nn as nn
from einops import rearrange, repeat


def rotate_half(x: torch.Tensor) -> torch.Tensor:
    x = rearrange(x, "... (d r) -> ... d r", r=2)
    x1, x2 = x.unbind(dim=-1)
    x = torch.stack((-x2, x1), dim=-1)
    return rearrange(x, "... d r -> ... (d r)")


class RotaryPositionalEmbedding1D(nn.Module):
    def __init__(
        self,
        head_dim: int,
    ):
        super().__init__()
        self.head_dim = head_dim
        self.base = 10000

    @lru_cache(maxsize=32)
    def precompute_freqs_cis_1d(self, pos_indices: torch.Tensor) -> torch.Tensor:
        freqs = 1.0 / (
            self.base
            ** (
                torch.arange(0, self.head_dim, 2)[: (self.head_dim // 2)].float()
                / self.head_dim
            )
        )
        freqs = freqs.to(pos_indices.device)
        freqs = torch.einsum("..., f -> ... f", pos_indices.float(), freqs)
        freqs = repeat(freqs, "... n -> ... (n r)", r=2)
        return freqs

    def forward(self, x: torch.Tensor, pos_indices: torch.Tensor) -> torch.Tensor:
        """1D RoPE.

        Args:
            x: [B, head, seq, head_dim]
            pos_indices: [seq,]
        Returns:
            query with the same shape as input.
        """
        freqs_cis = self.precompute_freqs_cis_1d(pos_indices)

        x_ = x.float()

        freqs_cis = freqs_cis.float().to(x.device)
        cos, sin = freqs_cis.cos(), freqs_cis.sin()
        cos, sin = rearrange(cos, "n d -> 1 1 n d"), rearrange(sin, "n d -> 1 1 n d")
        x_ = (x_ * cos) + (rotate_half(x_) * sin)

        return x_.type_as(x)


class VideoRopePosition3DEmb(nn.Module):
    def __init__(
        self,
        head_dim: int,
        len_h: int,
        len_w: int,
        len_t: int,
        h_extrapolation_ratio: float = 1.0,
        w_extrapolation_ratio: float = 1.0,
        t_extrapolation_ratio: float = 1.0,
    ):
        super().__init__()
        self.max_h = len_h
        self.max_w = len_w
        self.max_t = len_t
        dim = head_dim
        dim_h = dim // 6 * 2
        dim_w = dim_h
        dim_t = dim - 2 * dim_h
        assert dim == dim_h + dim_w + dim_t, (
            f"bad dim: {dim} != {dim_h} + {dim_w} + {dim_t}"
        )
        self._dim_h = dim_h
        self._dim_t = dim_t

        self.h_ntk_factor = h_extrapolation_ratio ** (dim_h / (dim_h - 2))
        self.w_ntk_factor = w_extrapolation_ratio ** (dim_w / (dim_w - 2))
        self.t_ntk_factor = t_extrapolation_ratio ** (dim_t / (dim_t - 2))

        self._is_initialized = False

    def cache_parameters(self) -> None:
        if self._is_initialized:
            return

        dim_h = self._dim_h
        dim_t = self._dim_t

        self.seq = torch.arange(max(self.max_h, self.max_w, self.max_t)).float().cuda()
        self.dim_spatial_range = (
            torch.arange(0, dim_h, 2)[: (dim_h // 2)].float().cuda() / dim_h
        )
        self.dim_temporal_range = (
            torch.arange(0, dim_t, 2)[: (dim_t // 2)].float().cuda() / dim_t
        )
        self._is_initialized = True

    def generate_embeddings(
        self,
        B_T_H_W_C: torch.Size,
        h_ntk_factor: Optional[float] = None,
        w_ntk_factor: Optional[float] = None,
        t_ntk_factor: Optional[float] = None,
    ):
        """
        Generate embeddings for the given input size.

        Args:
            B_T_H_W_C (torch.Size): Input tensor size (Batch, Time, Height, Width, Channels).
            h_ntk_factor (Optional[float], optional): Height NTK factor. If None, uses self.h_ntk_factor.
            w_ntk_factor (Optional[float], optional): Width NTK factor. If None, uses self.w_ntk_factor.
            t_ntk_factor (Optional[float], optional): Time NTK factor. If None, uses self.t_ntk_factor.

        Returns:
            Not specified in the original code snippet.
        """
        self.cache_parameters()

        h_ntk_factor = h_ntk_factor if h_ntk_factor is not None else self.h_ntk_factor
        w_ntk_factor = w_ntk_factor if w_ntk_factor is not None else self.w_ntk_factor
        t_ntk_factor = t_ntk_factor if t_ntk_factor is not None else self.t_ntk_factor

        h_theta = 10000.0 * h_ntk_factor
        w_theta = 10000.0 * w_ntk_factor
        t_theta = 10000.0 * t_ntk_factor

        h_spatial_freqs = 1.0 / (h_theta**self.dim_spatial_range)
        w_spatial_freqs = 1.0 / (w_theta**self.dim_spatial_range)
        temporal_freqs = 1.0 / (t_theta**self.dim_temporal_range)

        B, T, H, W, _ = B_T_H_W_C
        assert H <= self.max_h and W <= self.max_w, (
            f"Input dimensions (H={H}, W={W}) exceed the maximum dimensions (max_h={self.max_h}, max_w={self.max_w})"
        )
        freqs_h = torch.outer(self.seq[:H], h_spatial_freqs)
        freqs_w = torch.outer(self.seq[:W], w_spatial_freqs)

        freqs_t = torch.outer(self.seq[:T], temporal_freqs)

        freqs_T_H_W_D = torch.cat(
            [
                repeat(freqs_t, "t d -> t h w d", h=H, w=W),
                repeat(freqs_h, "h d -> t h w d", t=T, w=W),
                repeat(freqs_w, "w d -> t h w d", t=T, h=H),
            ],
            dim=-1,
        )

        return rearrange(freqs_T_H_W_D, "t h w d -> (t h w) d").float()

    @property
    def seq_dim(self):
        return 0



# SPDX-License-Identifier: Apache-2.0
# Written by Dr. Hicham Badri @Mobius Labs GmbH - 2025

import torch
import gc
import time
from tqdm import tqdm
from .core import GemLiteLinearTriton, DType, TORCH_TO_DTYPE
from .triton_kernels.utils import M_MAPPING
from .quant_utils import WeightQuantizerMXFP
from .triton_kernels.utils import IS_HIP

if IS_HIP:
    default_fp8 = torch.float8_e4m3fnuz  # AMD
    # default_fp8 = torch.float8_e5m2fnuz #AMD
    # default_fp8 = torch.float8_e5m2 #AMD - fp16 emulated
    default_post_scale = True
else:
    default_fp8 = torch.float8_e4m3fn  # Nvidia
    # default_fp8 = torch.float8_e5m2 #Nvidia
    default_post_scale = False


#################################################################################################
# Clean-up
def cleanup_linear(linear_layer, del_orig=True):
    if del_orig:
        for attr in ["weight", "bias", "weight_scale", "W_q", "meta"]:
            val = getattr(linear_layer, attr, None)
            if val is not None and hasattr(val, "__len__") and len(val) > 0:
                setattr(linear_layer, attr, None)
    torch.cuda.empty_cache()


# Replaces all linear layers with the corresponding processor
def patch_model(
    model,
    device,
    processor,
    skip_modules=["lm_head", "vision", "visual"],
    group_size=64,
):
    """
    Helper function to quantize the whole model from cpu device.
    The group_size parameter is only used in HQQ for W_nbits <=4, all the other configs do not use this parameter.
    """
    if hasattr(processor, "from_hqqlinear"):
        try:
            from hqq.core.quantize import HQQLinear, BaseQuantizeConfig
        except ImportException:
            print(
                "This processor requires the `hqq` package. Install it via `pip install hqq`."
            )
    else:

        class _NoHQQ:
            pass

        HQQLinear = _NoHQQ

    # Name modules
    for name, module in model.named_modules():
        module.name = name

    # Patching fct
    def _patching_fct(layer, device, skip_modules):
        layer = layer.to(device=device, non_blocking=True)
        if any(s in layer.name for s in skip_modules):
            return layer
        else:
            if hasattr(processor, "from_hqqlinear"):
                if isinstance(layer, torch.nn.Linear):
                    W_nbits = processor.W_nbits
                    quant_config = BaseQuantizeConfig(
                        nbits=W_nbits, group_size=group_size if W_nbits <= 4 else None
                    )
                    layer = HQQLinear(
                        layer,
                        quant_config=quant_config,
                        compute_dtype=layer.weight.dtype,
                        device=device,
                    )
                    layer = processor.from_hqqlinear(layer)
            else:
                layer = processor.from_linear(layer)

        return layer

    # Replaces linear layers
    def _patch_linearlayers(model, fct, device, skip_modules):
        for name, layer in model.named_children():
            if isinstance(layer, (torch.nn.Linear, HQQLinear)):
                setattr(model, name, fct(layer, device, skip_modules))
            else:
                _patch_linearlayers(layer, fct, device, skip_modules)

    # Apply patch
    _patch_linearlayers(model, _patching_fct, device, skip_modules)

    # Clean-up
    model = model.to(device=device)
    torch.cuda.empty_cache()
    gc.collect()

    return model


# 16-bit activations / 8-bit weigths
class A16W8:  # INT8/FP8 weight-only channel-wise
    def __init__(
        self, device="cuda:0", dtype=None, fp8=None, fp32_scale=True, post_scale=False
    ):
        self.device = device
        self.dtype = dtype
        self.fp8 = fp8
        self.fp32_scale = fp32_scale
        self.post_scale = post_scale

    def from_weights(self, weight, bias=None, scales=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        in_features, out_features = weight.shape[::-1]

        if scales is None:
            # Quantize
            if self.fp8:  # FP8
                w_dtype, min_val, max_val = (
                    self.fp8,
                    torch.finfo(self.fp8).min,
                    torch.finfo(self.fp8).max,
                )
            else:  # INT8
                w_dtype, min_val, max_val = (
                    torch.int8,
                    torch.iinfo(torch.int8).min,
                    torch.iinfo(torch.int8).max,
                )

            dtype = weight.dtype if (self.dtype is None) else self.dtype

            assert dtype in [
                torch.float16,
                torch.bfloat16,
                torch.float32,
            ], f"Invalid weight dtype, should be floating point, got {dtype}"

            gemlite_dtype = TORCH_TO_DTYPE[dtype]
            data_ptr = weight.data_ptr()
            weight = weight.to(dtype=torch.float32, copy=False, device=self.device)
            scales = weight.abs().amax(axis=1, keepdim=True) / max_val
            scales.clamp_(min=1e-6)
            W_q = (weight / scales).clamp_(min=min_val, max=max_val)
            if w_dtype.is_floating_point:
                W_q = W_q.to(w_dtype)
            else:
                W_q = W_q.round_().to(w_dtype)
            if data_ptr != weight.data_ptr():
                del weight
                torch.cuda.empty_cache()
        else:
            # Pre-Quantized
            assert weight.itemsize == 1, (
                f"Invalid weight.dtype, should be 8-bit (INT8 or FP8), got {weight.dtype}"
            )
            if self.dtype is None:
                dtype = (
                    scales.dtype
                    if scales.dtype in [torch.float16, torch.bfloat16]
                    else torch.float16
                )
            else:
                dtype = self.dtype
            W_q = weight.to(device=self.device)
            scales = scales.to(device=self.device)
            gemlite_dtype = TORCH_TO_DTYPE[dtype]

        # scales = scales.to(dtype=torch.float32 if self.fp32_scale else dtype)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            8,
            group_size=in_features,
            in_features=in_features,
            out_features=out_features,
            input_dtype=gemlite_dtype,
            output_dtype=gemlite_dtype,
        )

        gemlite_linear.pack(W_q, scales, zeros=None, bias=bias)

        if self.post_scale:
            # Post-scale
            gemlite_linear.W_group_mode = 0
            gemlite_linear.channel_scale_mode = 1
        else:
            # Pre-scaling
            gemlite_linear.W_group_mode = 2
            gemlite_linear.channel_scale_mode = 0

        return gemlite_linear

    def from_linear(self, linear_layer, del_orig=True):
        out_layer = self.from_weights(
            weight=linear_layer.weight, bias=linear_layer.bias
        )
        cleanup_linear(linear_layer, del_orig)
        return out_layer


# FP16 activations / Wn packed weights
class A16Wn:  # 8/4/2-bit weight-only as grouped "INT" / 8/4-bit as MXFP type
    def __init__(
        self,
        device="cuda:0",
        dtype=None,
        packing_bitwidth=None,
        post_scale=default_post_scale,
    ):
        self.post_scale = post_scale
        self.device = device
        self.dtype = dtype
        self.packing_bitwidth = packing_bitwidth
        self.quantizer_mx = None
        self.mx_fp8_dtype = default_fp8

    def from_weights(
        self, W_q, scales, zeros, W_nbits, group_size, bias=None, quant_type="INT"
    ):
        return self.from_weights_(
            W_q, scales, zeros, W_nbits, group_size, bias, quant_type
        )

    def from_weights_(
        self, W_q, scales, zeros, W_nbits, group_size, bias=None, quant_type="INT"
    ):
        assert quant_type in ["INT", "MXFP"], (
            f"Invalid quant_type. Got {quant_type}, valid values are INT, MXFP."
        )

        if quant_type == "MXFP":
            assert W_nbits in [8, 4], "Unsupported W_nbit for MXFP quant_dtype."
            assert group_size == 32, "group_size should 32 for MXFP."

        if isinstance(W_q, torch.nn.Parameter):
            W_q = W_q.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        if quant_type == "INT":
            dtype = scales.dtype if (self.dtype is None) else self.dtype
            scales = scales.to(dtype)

            assert scales.dtype in [
                torch.float16,
                torch.bfloat16,
                torch.float32,
            ], "Invalid scales.dtype, should floating point."

            gemlite_dtype = TORCH_TO_DTYPE[dtype]
            scales = scales.to(dtype)
            zeros = zeros.to(dtype)

        elif quant_type == "MXFP":
            if W_nbits == 8:
                assert W_q.dtype in [self.mx_fp8_dtype], (
                    f"Unsupported dtype of W_q. got {W_q.dtype}"
                )
            if W_nbits == 4:
                assert W_q.dtype in [torch.uint8], (
                    f"Unsupported dtype of W_q. got {W_q.dtype}"
                )

            dtype = torch.float16 if (self.dtype is None) else self.dtype
            if dtype == torch.float16:
                gemlite_dtype = DType.MXFP16
            elif dtype == torch.bfloat16:
                gemlite_dtype = DType.MXBF16
            else:
                raise Exception(
                    f"Unsupported dtype for MXFP. Got {dtype}, supported [torch.float16, torch.bfloat16]"
                )
            self.post_scale = False

            N, K = W_q.shape
            W_q, scales = W_q.view([N, K]), scales.view(N, K // group_size)  # .float()

        in_features, out_features = W_q.shape[::-1]

        W_q = W_q.to(self.device)
        scales = scales.to(device=self.device) if (scales is not None) else None
        zeros = zeros.to(device=self.device) if (zeros is not None) else None
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            W_nbits,
            group_size=group_size,
            in_features=in_features,
            out_features=out_features,
            input_dtype=gemlite_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=False,
        )

        gemlite_linear.pack(
            W_q, scales, zeros, bias=bias, packing_bitwidth=self.packing_bitwidth
        )

        if group_size == in_features and quant_type == "INT":
            if self.post_scale:
                gemlite_linear.W_group_mode = 1
                gemlite_linear.channel_scale_mode = 1
            else:
                gemlite_linear.W_group_mode = 3
                gemlite_linear.channel_scale_mode = 0

        return gemlite_linear

    def from_hqqlinear(self, hqq_layer, del_orig=True):
        assert hqq_layer.meta["axis"] == 1, "Only axis==1 is supported."

        self.device = hqq_layer.W_q.device

        W_nbits = hqq_layer.meta["nbits"]
        group_size = hqq_layer.meta["group_size"]
        if group_size is None:
            group_size = hqq_layer.in_features

        # Expects uint8 for Wn quantization!
        W_q = hqq_layer.unpack(dtype=torch.uint8).view(hqq_layer.meta["shape"])
        scales = hqq_layer.meta["scale"].clone()
        zeros = hqq_layer.meta["zero"].clone()
        bias = hqq_layer.bias.clone() if (hqq_layer.bias is not None) else None

        cleanup_linear(hqq_layer, del_orig)

        out_layer = self.from_weights_(
            W_q=W_q,
            scales=scales,
            zeros=zeros,
            W_nbits=W_nbits,
            group_size=group_size,
            bias=bias,
            quant_type="INT",
        )

        # Clean-up
        del W_q
        torch.cuda.empty_cache()

        return out_layer

    def mxfp_from_linear(self, linear_layer, W_nbits, del_orig=True):
        if self.quantizer_mx is None:
            self.quantizer_mx = WeightQuantizerMXFP(
                device=self.device, compute_dtype=linear_layer.weight.dtype
            )

        W = linear_layer.weight.data
        bias = linear_layer.bias.clone() if (linear_layer.bias is not None) else None
        group_size = 32
        N, K = W.shape
        if W_nbits == 8:
            W_q, scales = self.quantizer_mx.quantize_mxfp8(
                W, index=True, mx_fp8_dtype=self.mx_fp8_dtype
            )
        if W_nbits == 4:
            W_q, scales = self.quantizer_mx.quantize_mxfp4(W, index=True)
        W_q, scales = W_q.view([N, K]), scales.view(N, K // group_size)

        cleanup_linear(linear_layer, del_orig)

        out_layer = self.from_weights_(
            W_q=W_q,
            scales=scales,
            zeros=None,
            W_nbits=W_nbits,
            group_size=group_size,
            bias=bias,
            quant_type="MXFP",
        )

        # Clean-uo
        del W_q
        torch.cuda.empty_cache()
        return out_layer


# Alias
class A16W8_INT8(A16W8):
    def __init__(self, device="cuda:0", dtype=None, fp32_scale=True, post_scale=False):
        super().__init__(
            device=device,
            dtype=dtype,
            fp8=None,
            fp32_scale=fp32_scale,
            post_scale=post_scale,
        )


class A16W8_FP8(A16W8):
    def __init__(
        self,
        device="cuda:0",
        dtype=None,
        fp8=default_fp8,
        fp32_scale=True,
        post_scale=False,
    ):
        super().__init__(
            device=device,
            dtype=dtype,
            fp8=fp8,
            fp32_scale=fp32_scale,
            post_scale=post_scale,
        )


class A16Wn_HQQ_INT(A16Wn):
    def __init__(self, device="cuda:0", dtype=None, W_nbits=None):
        super().__init__(device=device, dtype=dtype)
        self.W_nbits = W_nbits

    def from_weights(self, W_q, scales, zeros, bias=None):
        group_size = W_q.numel() // scales.numel()
        return super().from_weights(
            W_q=W_q,
            scales=scales,
            zeros=zeros,
            W_nbits=self.W_nbits,
            group_size=group_size,
            bias=bias,
            quant_type="INT",
        )


class A16W8_HQQ_INT(A16Wn_HQQ_INT):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__(device=device, dtype=dtype, W_nbits=8)


class A16W4_HQQ_INT(A16Wn_HQQ_INT):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__(device=device, dtype=dtype, W_nbits=4)


class A16W2_HQQ_INT(A16Wn_HQQ_INT):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__(device=device, dtype=dtype, W_nbits=2)


class A16W1_HQQ_INT(A16Wn_HQQ_INT):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__(device=device, dtype=dtype, W_nbits=1)


class A16Wn_MXFP(A16Wn):
    def __init__(self, device="cuda:0", dtype=None, W_nbits=None):
        super().__init__(device=device, dtype=dtype)
        self.W_nbits = W_nbits

    def from_weights(self, W_q, scales, bias=None):
        group_size = W_q.numel() // scales.numel()
        return super().from_weights(
            W_q=W_q,
            scales=scales,
            zeros=None,
            W_nbits=self.W_nbits,
            group_size=group_size,
            bias=bias,
            quant_type="MXFP",
        )

    def from_linear(self, linear_layer, del_orig=True):
        return super().mxfp_from_linear(
            linear_layer=linear_layer, W_nbits=self.W_nbits, del_orig=del_orig
        )


class A16W8_MXFP(A16Wn_MXFP):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__(device=device, dtype=dtype, W_nbits=8)


class A16W4_MXFP(A16Wn_MXFP):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__(device=device, dtype=dtype, W_nbits=4)


#################################################################################################
# 8-bit dynamic activations / 8-bit weights
#################################################################################################
class A8W8_dynamic:
    def __init__(self, device="cuda:0", dtype=None, fp8=False, fp32_scale=True):
        self.device = device
        self.dtype = dtype
        self.fp8 = fp8
        self.fp32_scale = fp32_scale

    def from_weights(self, weight, bias=None, scales=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        if self.fp8:  # FP8
            w_dtype, input_dtype, min_val, max_val = (
                self.fp8,
                TORCH_TO_DTYPE[self.fp8],
                torch.finfo(self.fp8).min,
                torch.finfo(self.fp8).max,
            )
        else:  # INT8
            w_dtype, input_dtype, min_val, max_val = (
                torch.int8,
                DType.INT8,
                torch.iinfo(torch.int8).min,
                torch.iinfo(torch.int8).max,
            )

        in_features, out_features = weight.shape[::-1]

        if scales is None:
            # Quantize
            dtype = weight.dtype if (self.dtype is None) else self.dtype

            assert dtype in [
                torch.float16,
                torch.bfloat16,
                torch.float32,
            ], "Invalid weight dtype, should be floating point."

            gemlite_dtype = TORCH_TO_DTYPE[dtype]
            data_ptr = weight.data_ptr()
            weight = weight.to(dtype=torch.float32, copy=False, device=self.device)
            scales = weight.abs().amax(axis=1, keepdim=True) / max_val
            scales.clamp_(min=1e-6)
            W_q = (weight / scales).clamp_(min=min_val, max=max_val)
            if w_dtype.is_floating_point:
                W_q = W_q.to(w_dtype)
            else:
                W_q = W_q.round_().to(w_dtype)
            if data_ptr != weight.data_ptr():
                del weight
                torch.cuda.empty_cache()
        else:
            # Pre-Quantized
            assert weight.dtype.itemsize == 1, "Invalid weight.dtype, should be 8-bit."
            if self.dtype is None:
                dtype = (
                    scales.dtype
                    if scales.dtype in [torch.float16, torch.bfloat16]
                    else torch.float16
                )
            else:
                dtype = self.dtype
            W_q = weight.to(device=self.device)
            scales = scales.to(device=self.device)
            gemlite_dtype = TORCH_TO_DTYPE[dtype]

        scales = scales.to(torch.float32 if self.fp32_scale else dtype)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            8,
            group_size=in_features,
            in_features=in_features,
            out_features=out_features,
            input_dtype=input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        gemlite_linear.pack(W_q, scales, zeros=None, bias=bias)

        # Post-scaling
        gemlite_linear.W_group_mode = 0
        gemlite_linear.channel_scale_mode = 3
        return gemlite_linear

    def from_linear(self, linear_layer, del_orig=True):
        out_layer = self.from_weights(
            weight=linear_layer.weight, bias=linear_layer.bias
        )
        cleanup_linear(linear_layer, del_orig)
        return out_layer


class A8W8_int8_dynamic(A8W8_dynamic):
    def __init__(self, device="cuda:0", dtype=None):
        super().__init__()
        self.device = device
        self.dtype = dtype
        self.fp8 = False


A8W8_INT8_dynamic = A8W8_int8_dynamic


class A8W8_fp8_dynamic(A8W8_dynamic):
    def __init__(self, device="cuda:0", dtype=None, fp8=default_fp8):
        super().__init__()
        self.device = device
        self.dtype = dtype
        self.fp8 = fp8


A8W8_FP8_dynamic = A8W8_fp8_dynamic


#################################################################################################
# FP8 dynamic activations / W4 packed weights
#################################################################################################
class A8Wn_HQQ_INT_dynamic(A16Wn):
    def __init__(
        self,
        device="cuda:0",
        packing_bitwidth=None,
        dtype=None,
        post_scale=default_post_scale,
        fp8=default_fp8,
        fp32_scale=False,
        W_nbits=None,
    ):
        assert W_nbits is not None, (
            "W_nbits argument should be eitehr 8 or 4, not None)."
        )
        super().__init__()
        self.post_scale = post_scale
        self.device = device
        self.dtype = dtype
        self.packing_bitwidth = packing_bitwidth
        self.fp8 = fp8
        self.fp32_scale = fp32_scale
        self.W_nbits = W_nbits

    def from_weights(self, W_q, scales, zeros, bias=None):
        group_size = W_q.numel() // scales.numel()
        return self.from_weights_(
            W_q=W_q,
            scales=scales,
            zeros=zeros,
            W_nbits=self.W_nbits,
            group_size=group_size,
            bias=bias,
        )

    def from_weights_(self, W_q, scales, zeros, W_nbits, group_size, bias=None):
        if isinstance(W_q, torch.nn.Parameter):
            W_q = W_q.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        if self.dtype is None:
            dtype = (
                scales.dtype
                if scales.dtype in [torch.float16, torch.bfloat16]
                else torch.float16
            )
        else:
            dtype = self.dtype

        assert dtype in [torch.float16, torch.bfloat16, torch.float32], (
            "Invalid scales.dtype, should be floating point."
        )
        gemlite_dtype = TORCH_TO_DTYPE[dtype]
        input_dtype = TORCH_TO_DTYPE[self.fp8]

        W_q = W_q.to(self.device)
        scales = (
            scales.to(
                device=self.device, dtype=torch.float32 if self.fp32_scale else dtype
            )
            if (scales is not None)
            else None
        )
        zeros = (
            zeros.to(device=self.device, dtype=dtype) if (zeros is not None) else None
        )
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        in_features, out_features = W_q.shape[::-1]

        gemlite_linear = GemLiteLinearTriton(
            W_nbits,
            group_size=group_size,
            in_features=in_features,
            out_features=out_features,
            input_dtype=input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        gemlite_linear.pack(
            W_q,
            scales,
            zeros,
            bias=bias,
            packing_bitwidth=self.packing_bitwidth,
            fma_mode=False,
        )

        if group_size == in_features:
            if self.post_scale:
                gemlite_linear.W_group_mode = 1
                gemlite_linear.channel_scale_mode = 3
            else:
                gemlite_linear.W_group_mode = 3
                gemlite_linear.channel_scale_mode = 2

        return gemlite_linear

    def from_hqqlinear(self, hqq_layer, del_orig=True):
        assert hqq_layer.meta["axis"] == 1, "Only axis==1 is supported."

        self.device = hqq_layer.W_q.device

        W_nbits = hqq_layer.meta["nbits"]
        group_size = hqq_layer.meta["group_size"]
        if group_size is None:
            group_size = hqq_layer.in_features

        W_q = hqq_layer.unpack(dtype=torch.uint8).view(
            hqq_layer.meta["shape"]
        )  # Expects uint8 for Wn quantization!
        scales = hqq_layer.meta["scale"].clone()
        zeros = hqq_layer.meta["zero"].clone()
        bias = hqq_layer.bias.clone() if (hqq_layer.bias is not None) else None

        cleanup_linear(hqq_layer, del_orig)

        out_layer = self.from_weights_(
            W_q=W_q,
            scales=scales,
            zeros=zeros,
            W_nbits=W_nbits,
            group_size=group_size,
            bias=bias,
        )

        # Clean-up
        del W_q
        torch.cuda.empty_cache()

        return out_layer


class A8W4_HQQ_INT_dynamic(A8Wn_HQQ_INT_dynamic):
    def __init__(
        self,
        device="cuda:0",
        packing_bitwidth=None,
        dtype=None,
        post_scale=default_post_scale,
        fp8=default_fp8,
        fp32_scale=False,
    ):
        super().__init__(
            device=device,
            packing_bitwidth=packing_bitwidth,
            dtype=dtype,
            post_scale=post_scale,
            fp8=fp8,
            fp32_scale=fp32_scale,
            W_nbits=4,
        )


class A8W2_HQQ_INT_dynamic(A8Wn_HQQ_INT_dynamic):
    def __init__(
        self,
        device="cuda:0",
        packing_bitwidth=None,
        dtype=None,
        post_scale=default_post_scale,
        fp8=default_fp8,
        fp32_scale=False,
    ):
        super().__init__(
            device=device,
            packing_bitwidth=packing_bitwidth,
            dtype=dtype,
            post_scale=post_scale,
            fp8=fp8,
            fp32_scale=fp32_scale,
            W_nbits=2,
        )


#################################################################################################
# MXFP / NVFP dtypes with dynamic quant activations
#################################################################################################
class A8W8_MXFP_dynamic:
    def __init__(self, device="cuda:0", dtype=None, post_scale=True, fp8=default_fp8):
        self.device = device
        self.dtype = dtype
        self.mx_fp8_dtype = fp8
        self.quantizer_mx = None
        self.post_scale = post_scale
        self.W_nbits = 8

    def from_weights(self, weight, bias=None, scales=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        in_features, out_features = weight.shape[::-1]

        assert scales is not None, (
            "Scales parameter cannot be None. Use from_linear() call to pre-quantize the weights."
        )

        # Pre-Quantized
        assert weight.is_floating_point() and weight.itemsize == 1, (
            f"Invalid weight.dtype, should be an MXPF8 dype, got {weight.dtype}."
        )
        assert scales.dtype in [torch.float8_e8m0fnu, torch.uint8], (
            f"Invalid scales.dtype, should be e8m0 / view(uint8), got {scales.dtype}."
        )
        assert self.dtype is not None, (
            "Input dtype should be either torch.float16 or torch.bfloat16, not None."
        )
        dtype = self.dtype
        input_dtype = DType.MXFP8
        gemlite_dtype = TORCH_TO_DTYPE[dtype]
        group_size = 32

        W_q = weight.to(device=self.device)
        scales = scales.to(device=self.device)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            self.W_nbits,
            group_size=group_size,
            in_features=in_features,
            out_features=out_features,
            input_dtype=input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        gemlite_linear.pack(W_q, scales, zeros=None, bias=bias)

        # If post_scale==False, it will use mxfp8 microscales for the activations, otherwise channelwise post-scaling is used
        gemlite_linear.W_group_mode = 0
        gemlite_linear.channel_scale_mode = 2 if self.post_scale else 4
        return gemlite_linear

    def from_linear(self, linear_layer, del_orig=True):
        if self.quantizer_mx is None:
            self.quantizer_mx = WeightQuantizerMXFP(
                device=self.device, compute_dtype=linear_layer.weight.dtype
            )

        W = linear_layer.weight.data
        bias = linear_layer.bias.clone() if (linear_layer.bias is not None) else None
        group_size = 32
        N, K = W.shape
        if self.W_nbits == 8:
            W_q, scales = self.quantizer_mx.quantize_mxfp8(
                W, index=True, mx_fp8_dtype=self.mx_fp8_dtype
            )
        if self.W_nbits == 4:
            W_q, scales = self.quantizer_mx.quantize_mxfp4(W, index=True)
        W_q, scales = W_q.view([N, K]), scales.view(N, K // group_size)

        cleanup_linear(linear_layer, del_orig)

        out_layer = self.from_weights(weight=W_q, scales=scales, bias=bias)

        # Clean-uo
        del W_q
        torch.cuda.empty_cache()
        return out_layer


class A8Wn_MXFP_dynamic:
    def __init__(
        self,
        device="cuda:0",
        dtype=None,
        post_scale=True,
        fp8=default_fp8,
        W_nbits=None,
    ):
        assert W_nbits is not None, (
            "W_nbits argument should be eitehr 8 or 4, not None)."
        )
        self.device = device
        self.dtype = dtype
        self.mx_fp8_dtype = fp8
        self.quantizer_mx = None
        self.post_scale = post_scale
        self.W_nbits = W_nbits

    def from_weights(self, weight, bias=None, scales=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        in_features, out_features = weight.shape[::-1]

        assert scales is not None, (
            "Scales parameter cannot be None. Use from_linear() call to pre-quantize the weights."
        )

        # Pre-Quantized
        if self.W_nbits == 8:
            assert weight.is_floating_point() and weight.itemsize == 1, (
                f"Invalid weight.dtype, should be an MXPF8 valid dtype, got {weight.dtype}."
            )
        if self.W_nbits == 4:
            assert weight.dtype in [torch.uint8], (
                f"Invalid weight.dtype, should be an MXPF8 (FP8 dtype) valid dtype, got {weight.dtype}."
            )
        assert scales.dtype in [torch.float8_e8m0fnu, torch.uint8], (
            f"Invalid scales.dtype, should be e8m0 / view(uint8), got {scales.dtype}."
        )
        assert self.dtype is not None, (
            "Input dtype should be either torch.float16 or torch.bfloat16, not None."
        )
        dtype = self.dtype
        input_dtype = DType.MXFP8
        gemlite_dtype = TORCH_TO_DTYPE[dtype]
        group_size = 32

        W_q = weight.to(device=self.device)
        scales = scales.to(device=self.device)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            self.W_nbits,
            group_size=group_size,
            in_features=in_features,
            out_features=out_features,
            input_dtype=input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        gemlite_linear.pack(W_q, scales, zeros=None, bias=bias)

        # If post_scale==False, it will use mxfp8 microscales for the activations, otherwise channelwise post-scaling is used
        gemlite_linear.W_group_mode = 0
        gemlite_linear.channel_scale_mode = 2 if self.post_scale else 4
        return gemlite_linear

    def from_linear(self, linear_layer, del_orig=True):
        if self.quantizer_mx is None:
            self.quantizer_mx = WeightQuantizerMXFP(
                device=self.device, compute_dtype=linear_layer.weight.dtype
            )

        W = linear_layer.weight.data
        bias = linear_layer.bias.clone() if (linear_layer.bias is not None) else None
        group_size = 32
        N, K = W.shape
        if self.W_nbits == 8:
            W_q, scales = self.quantizer_mx.quantize_mxfp8(
                W, index=True, mx_fp8_dtype=self.mx_fp8_dtype
            )
        if self.W_nbits == 4:
            W_q, scales = self.quantizer_mx.quantize_mxfp4(W, index=True)
        W_q, scales = W_q.view([N, K]), scales.view(N, K // group_size)

        cleanup_linear(linear_layer, del_orig)

        out_layer = self.from_weights(weight=W_q, scales=scales, bias=bias)

        # Clean-uo
        del W_q
        torch.cuda.empty_cache()
        return out_layer


class A8W8_MXFP_dynamic(A8Wn_MXFP_dynamic):
    def __init__(self, device="cuda:0", dtype=None, post_scale=True, fp8=default_fp8):
        super().__init__(
            device=device, dtype=dtype, post_scale=post_scale, fp8=fp8, W_nbits=8
        )


class A8W4_MXFP_dynamic(A8Wn_MXFP_dynamic):
    def __init__(self, device="cuda:0", dtype=None, post_scale=True, fp8=default_fp8):
        super().__init__(
            device=device, dtype=dtype, post_scale=post_scale, fp8=fp8, W_nbits=4
        )


class A4W4_MXFP_dynamic:
    def __init__(self, device="cuda:0", dtype=None):
        self.device = device
        self.dtype = dtype
        self.quantizer_mx = None
        self.W_nbits = 4
        self.group_size = 32
        self.input_dtype = DType.MXFP4

    def from_weights(self, weight, bias=None, scales=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        in_features, out_features = weight.shape[::-1]

        assert scales is not None, (
            "Scales parameter cannot be None. Use from_linear() call to pre-quantize the weights."
        )

        # Pre-Quantized
        assert weight.dtype in [torch.uint8], (
            f"Invalid weight.dtype, should be an MXPF8 valid dtype, got {weight.dtype}."
        )
        assert scales.dtype in [torch.float8_e8m0fnu, torch.uint8], (
            f"Invalid scales.dtype, should be e8m0 / view(uint8), got {scales.dtype}."
        )
        assert self.dtype is not None, (
            "Input dtype should be either torch.float16 or torch.bfloat16, not None."
        )
        assert self.group_size == 32, (
            f"Only group_size=16 is supported for MXFP4, got {self.group_size}"
        )

        dtype = self.dtype
        gemlite_dtype = TORCH_TO_DTYPE[dtype]

        W_q = weight.to(device=self.device)
        scales = scales.to(device=self.device)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            self.W_nbits,
            group_size=self.group_size,
            in_features=in_features,
            out_features=out_features,
            input_dtype=self.input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        gemlite_linear.pack(W_q, scales, zeros=None, bias=bias)
        gemlite_linear.W_group_mode = 0
        gemlite_linear.channel_scale_mode = 4
        return gemlite_linear

    def from_linear(self, linear_layer, del_orig=True):
        if self.quantizer_mx is None:
            self.quantizer_mx = WeightQuantizerMXFP(
                device=self.device, compute_dtype=linear_layer.weight.dtype
            )

        W = linear_layer.weight.data
        bias = linear_layer.bias.clone() if (linear_layer.bias is not None) else None
        N, K = W.shape
        W_q, scales = self.quantizer_mx.quantize_mxfp4(W, index=True)
        W_q, scales = W_q.view([N, K]), scales.view(N, K // self.group_size)

        cleanup_linear(linear_layer, del_orig)

        out_layer = self.from_weights(weight=W_q, scales=scales, bias=bias)

        # Clean-uo
        del W_q
        torch.cuda.empty_cache()
        return out_layer


class A4W4_NVFP_dynamic:
    def __init__(self, device="cuda:0", dtype=None):
        self.device = device
        self.dtype = dtype
        self.quantizer_mx = None
        self.W_nbits = 4
        self.group_size = 16
        self.input_dtype = DType.NVFP4

    def from_weights(self, weight, bias=None, scales=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        in_features, out_features = weight.shape[::-1]

        assert scales is not None, (
            "Scales parameter cannot be None. Use from_linear() call to pre-quantize the weights."
        )

        # Pre-Quantized
        assert weight.dtype in [torch.uint8], (
            f"Invalid weight.dtype, should be an MXPF8 dtype, got {weight.dtype}."
        )
        assert scales.dtype in [torch.float8_e4m3fn], (
            f"Invalid scales.dtype, should be float8_e4m3fn, got {scales.dtype}."
        )
        assert self.dtype is not None, (
            "Input dtype should be either torch.float16 or torch.bfloat16, not None."
        )
        assert self.group_size == 16, (
            f"Only group_size=16 is supported for NVFP4, got {self.group_size}"
        )

        dtype = self.dtype
        gemlite_dtype = TORCH_TO_DTYPE[dtype]

        W_q = weight.to(device=self.device)
        scales = scales.to(device=self.device)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear = GemLiteLinearTriton(
            self.W_nbits,
            group_size=self.group_size,
            in_features=in_features,
            out_features=out_features,
            input_dtype=self.input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        gemlite_linear.pack(W_q, scales, zeros=None, bias=bias)
        gemlite_linear.W_group_mode = 0
        gemlite_linear.channel_scale_mode = 4
        return gemlite_linear

    def from_linear(self, linear_layer, del_orig=True):
        if self.quantizer_mx is None:
            self.quantizer_mx = WeightQuantizerMXFP(
                device=self.device, compute_dtype=linear_layer.weight.dtype
            )

        W = linear_layer.weight.data
        bias = linear_layer.bias.clone() if (linear_layer.bias is not None) else None
        N, K = W.shape
        W_q, scales = self.quantizer_mx.quantize_nvfp4(W, index=True)
        W_q, scales = W_q.view([N, K]), scales.view(N, K // self.group_size)
        cleanup_linear(linear_layer, del_orig)

        out_layer = self.from_weights(weight=W_q, scales=scales, bias=bias)

        # Clean-uo
        del W_q
        torch.cuda.empty_cache()
        return out_layer


#################################################################################################
# BitNet
#################################################################################################
class A16W158_INT:
    def __init__(self, device="cuda:0", dtype=None, fp32_scale=True):
        self.device = device
        self.dtype = dtype
        self.fp32_scale = fp32_scale

    def from_weights(self, weight, weight_scale, bias=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        dtype = weight.dtype if (self.dtype is None) else self.dtype

        assert dtype in [
            torch.float16,
            torch.bfloat16,
            torch.float32,
        ], "Invalid weight.dtype, should be floating point."

        weight = weight.to(dtype=dtype, device=self.device)
        W_q = (weight + 1).to(torch.uint8)
        bias = bias if bias is not None else None
        dtype = weight.dtype
        gemlite_dtype = TORCH_TO_DTYPE[dtype]
        out_features = W_q.shape[0]
        in_features = W_q.shape[1]
        scales = (
            torch.ones((out_features, 1), dtype=torch.float32) * weight_scale.item()
        )

        gemlite_linear = GemLiteLinearTriton(
            2,
            group_size=in_features,
            in_features=in_features,
            out_features=out_features,
            input_dtype=gemlite_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=False,
        )

        scales = scales.to(dtype=torch.float32 if self.fp32_scale else dtype)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear.pack(W_q, scales=scales, zeros=1, bias=bias)
        # post-scale
        gemlite_linear.W_group_mode = 1  # shift only
        gemlite_linear.channel_scale_mode = 1  # weight-only
        return gemlite_linear

    def from_bitlinear(self, linear_layer, del_orig=True):
        out_layer = self.from_weights(
            weight=linear_layer.weight,
            weight_scale=linear_layer.weight_scale,
            bias=linear_layer.bias,
        )
        cleanup_linear(linear_layer, del_orig)
        return out_layer


class A8W158_INT_dynamic:
    def __init__(self, device="cuda:0", dtype=None, fp32_scale=True):
        self.device = device
        self.dtype = dtype
        self.fp32_scale = fp32_scale

    def from_weights(self, weight, weight_scale, bias=None):
        if isinstance(weight, torch.nn.Parameter):
            weight = weight.data
        if isinstance(bias, torch.nn.Parameter):
            bias = bias.data

        dtype = weight.dtype if (self.dtype is None) else self.dtype

        assert dtype in [
            torch.float16,
            torch.bfloat16,
            torch.float32,
        ], "Invalid weight.dtype, should be floating point."

        weight = weight.to(device=self.device, dtype=dtype)
        W_q = (weight + 1).to(torch.uint8)
        dtype = weight.dtype
        gemlite_dtype = TORCH_TO_DTYPE[dtype]
        out_features = W_q.shape[0]
        in_features = W_q.shape[1]
        scales = (
            torch.ones((out_features, 1), dtype=torch.float32, device=self.device)
            * weight_scale.item()
        )
        input_dtype = DType.INT8

        gemlite_linear = GemLiteLinearTriton(
            2,
            group_size=in_features,
            in_features=in_features,
            out_features=out_features,
            input_dtype=input_dtype,
            output_dtype=gemlite_dtype,
            scaled_activations=True,
        )

        scales = scales.to(dtype=torch.float32 if self.fp32_scale else dtype)
        bias = bias.to(device=self.device, dtype=dtype) if (bias is not None) else None

        gemlite_linear.pack(W_q, scales=scales, zeros=1, bias=bias)

        # post-scale
        gemlite_linear.W_group_mode = 1  # shift only
        gemlite_linear.channel_scale_mode = 3  # activations + weight

        return gemlite_linear

    def from_bitlinear(self, linear_layer, del_orig=True):
        out_layer = self.from_weights(
            weight=linear_layer.weight,
            weight_scale=linear_layer.weight_scale,
            bias=linear_layer.bias,
        )
        cleanup_linear(linear_layer, del_orig)
        return out_layer


#################################################################################################
# Warm-up function
#################################################################################################
default_batch_sizes = sorted(list(set(M_MAPPING)))[::-1]


def warmup(
    processor,
    shapes: list[tuple],
    batch_sizes: list[int] = default_batch_sizes,
    group_size: int = 64,
    dtype: torch.dtype = torch.float16,
):
    """
    processor = A8W8_INT8_dynamic()
    warmup(processor, shapes=[(4096, 4096), (4096, 2048)], batch_sizes=[1, 4, 8, ..])
    `group_size` is only used for hqq quantization.
    """

    if hasattr(processor, "from_hqqlinear"):
        try:
            from hqq.core.quantize import HQQLinear, BaseQuantizeConfig
        except ImportException:
            print(
                "This processor requires the `hqq` package. Install it via `pip install hqq`."
            )
    else:

        class _NoHQQ:
            pass

        HQQLinear = _NoHQQ

    device = torch.device(torch.cuda.current_device())

    for shape in shapes:
        out_features, in_features = shape

        linear = torch.nn.Linear(
            in_features=in_features,
            out_features=out_features,
            bias=False,
            dtype=dtype,
            device=device,
        )

        if hasattr(processor, "from_hqqlinear"):
            W_nbits = processor.W_nbits
            quant_config = BaseQuantizeConfig(
                nbits=W_nbits, group_size=group_size, axis=1
            )
            linear = HQQLinear(
                linear, quant_config=quant_config, compute_dtype=dtype, device=device
            )
            gemlite_linear = processor.from_hqqlinear(linear)
        else:
            gemlite_linear = processor.from_linear(linear)

        for batch_size in tqdm(batch_sizes):
            _ = gemlite_linear(
                torch.randn((batch_size, in_features), dtype=dtype, device=device)
                / 100.0
            )
            torch.cuda.synchronize()

        torch.cuda.empty_cache()
        gc.collect()
        time.sleep(1)



# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
import torch
import torch.nn as nn
from einops import rearrange

from ..utils import (
    RotaryPositionalEmbedding1D,
    normalize_and_scale,
)
from ..kernels.attn import attention


class SingleStreamAttention(nn.Module):
    def __init__(
        self,
        dim: int,
        encoder_hidden_states_dim: int,
        num_heads: int,
        qkv_bias: bool,
        qk_norm: bool,
        norm_layer: nn.Module,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        eps: float = 1e-6,
    ) -> None:
        super().__init__()
        assert dim % num_heads == 0, "dim should be divisible by num_heads"
        self.dim = dim
        self.encoder_hidden_states_dim = encoder_hidden_states_dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.scale = self.head_dim**-0.5
        self.qk_norm = qk_norm

        self.q_linear = nn.Linear(dim, dim, bias=qkv_bias)

        self.q_norm = norm_layer(self.head_dim, eps=eps) if qk_norm else nn.Identity()
        self.k_norm = norm_layer(self.head_dim, eps=eps) if qk_norm else nn.Identity()

        self.attn_drop = nn.Dropout(attn_drop)
        self.proj = nn.Linear(dim, dim)
        self.proj_drop = nn.Dropout(proj_drop)

        self.kv_linear = nn.Linear(encoder_hidden_states_dim, dim * 2, bias=qkv_bias)

        self.add_q_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()
        self.add_k_norm = norm_layer(self.head_dim) if qk_norm else nn.Identity()

    def forward(
        self,
        x: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        shape=None,
        enable_sp=False,
        kv_seq=None,
        visual_seqlen=None,
    ) -> torch.Tensor:

        N_t, N_h, N_w = shape
        if not enable_sp:
            x = rearrange(x, "B (N_t S) C -> (B N_t) S C", N_t=N_t)

        # get q for hidden_state
        B, N, C = x.shape
        q = self.q_linear(x)
        q_shape = (B, N, self.num_heads, self.head_dim)
        q = q.view(q_shape).permute((0, 2, 1, 3))

        if self.qk_norm:
            q = self.q_norm(q)

        # get kv from encoder_hidden_states
        _, N_a, _ = encoder_hidden_states.shape
        encoder_kv = self.kv_linear(encoder_hidden_states)
        encoder_kv_shape = (B, N_a, 2, self.num_heads, self.head_dim)
        encoder_kv = encoder_kv.view(encoder_kv_shape).permute((2, 0, 3, 1, 4))
        encoder_k, encoder_v = encoder_kv.unbind(0)

        if self.qk_norm:
            encoder_k = self.add_k_norm(encoder_k)

        q = rearrange(q, "B H M K -> B M H K")
        encoder_k = rearrange(encoder_k, "B H M K -> B M H K")
        encoder_v = rearrange(encoder_v, "B H M K -> B M H K")

        x = attention(q, encoder_k, encoder_v)
        x = rearrange(x, "B M H K -> B H M K")

        # linear transform
        x_output_shape = (B, N, C)
        x = x.transpose(1, 2)
        x = x.reshape(x_output_shape)
        x = self.proj(x)
        x = self.proj_drop(x)

        if not enable_sp:
            # reshape x to origin shape
            x = rearrange(x, "(B N_t) S C -> B (N_t S) C", N_t=N_t)
        # x = torch.concat([x, x_a], dim=1)
        return x


class SingleStreamMutiAttention(SingleStreamAttention):
    def __init__(
        self,
        dim: int,
        encoder_hidden_states_dim: int,
        num_heads: int,
        qkv_bias: bool,
        qk_norm: bool,
        norm_layer: nn.Module,
        attn_drop: float = 0.0,
        proj_drop: float = 0.0,
        eps: float = 1e-6,
        class_range: int = 24,
        class_interval: int = 4,
    ) -> None:
        super().__init__(
            dim=dim,
            encoder_hidden_states_dim=encoder_hidden_states_dim,
            num_heads=num_heads,
            qkv_bias=qkv_bias,
            qk_norm=qk_norm,
            norm_layer=norm_layer,
            attn_drop=attn_drop,
            proj_drop=proj_drop,
            eps=eps,
        )
        self.class_interval = class_interval
        self.class_range = class_range
        self.rope_h1 = (0, self.class_interval)
        self.rope_h2 = (self.class_range - self.class_interval, self.class_range)
        self.rope_bak = int(self.class_range // 2)

        self.rope_1d = RotaryPositionalEmbedding1D(self.head_dim)

    def forward(
        self,
        x: torch.Tensor,
        encoder_hidden_states: torch.Tensor,
        shape=None,
        x_ref_attn_map=None,
        human_num=None,
    ) -> torch.Tensor:

        encoder_hidden_states = encoder_hidden_states.squeeze(0)
        if human_num == 1:
            return super().forward(x, encoder_hidden_states, shape)

        N_t, _, _ = shape
        x = rearrange(x, "B (N_t S) C -> (B N_t) S C", N_t=N_t)

        # get q for hidden_state
        B, N, C = x.shape
        q = self.q_linear(x)
        q_shape = (B, N, self.num_heads, self.head_dim)
        q = q.view(q_shape).permute((0, 2, 1, 3))

        if self.qk_norm:
            q = self.q_norm(q)

        max_values = x_ref_attn_map.max(1).values[:, None, None]
        min_values = x_ref_attn_map.min(1).values[:, None, None]
        max_min_values = torch.cat([max_values, min_values], dim=2)

        human1_max_value, human1_min_value = (
            max_min_values[0, :, 0].max(),
            max_min_values[0, :, 1].min(),
        )
        human2_max_value, human2_min_value = (
            max_min_values[1, :, 0].max(),
            max_min_values[1, :, 1].min(),
        )

        human1 = normalize_and_scale(
            x_ref_attn_map[0],
            (human1_min_value, human1_max_value),
            (self.rope_h1[0], self.rope_h1[1]),
        )
        human2 = normalize_and_scale(
            x_ref_attn_map[1],
            (human2_min_value, human2_max_value),
            (self.rope_h2[0], self.rope_h2[1]),
        )
        back = torch.full(
            (x_ref_attn_map.size(1),), self.rope_bak, dtype=human1.dtype
        ).to(human1.device)
        max_indices = x_ref_attn_map.argmax(dim=0)
        normalized_map = torch.stack([human1, human2, back], dim=1)
        normalized_pos = normalized_map[range(x_ref_attn_map.size(1)), max_indices]  # N

        q = rearrange(q, "(B N_t) H S C -> B H (N_t S) C", N_t=N_t)
        q = self.rope_1d(q, normalized_pos)
        q = rearrange(q, "B H (N_t S) C -> (B N_t) H S C", N_t=N_t)

        _, N_a, _ = encoder_hidden_states.shape
        encoder_kv = self.kv_linear(encoder_hidden_states)
        encoder_kv_shape = (B, N_a, 2, self.num_heads, self.head_dim)
        encoder_kv = encoder_kv.view(encoder_kv_shape).permute((2, 0, 3, 1, 4))
        encoder_k, encoder_v = encoder_kv.unbind(0)

        if self.qk_norm:
            encoder_k = self.add_k_norm(encoder_k)

        per_frame = torch.zeros(N_a, dtype=encoder_k.dtype).to(encoder_k.device)
        per_frame[: per_frame.size(0) // 2] = (self.rope_h1[0] + self.rope_h1[1]) / 2
        per_frame[per_frame.size(0) // 2 :] = (self.rope_h2[0] + self.rope_h2[1]) / 2
        encoder_pos = torch.concat([per_frame] * N_t, dim=0)
        encoder_k = rearrange(encoder_k, "(B N_t) H S C -> B H (N_t S) C", N_t=N_t)
        encoder_k = self.rope_1d(encoder_k, encoder_pos)
        encoder_k = rearrange(encoder_k, "B H (N_t S) C -> (B N_t) H S C", N_t=N_t)

        q = rearrange(q, "B H M K -> B M H K")
        encoder_k = rearrange(encoder_k, "B H M K -> B M H K")
        encoder_v = rearrange(encoder_v, "B H M K -> B M H K")
        x = attention(q=q, k=encoder_k, v=encoder_v)
        x = rearrange(x, "B M H K -> B H M K")

        # linear transform
        x_output_shape = (B, N, C)
        x = x.transpose(1, 2)
        x = x.reshape(x_output_shape)
        x = self.proj(x)
        x = self.proj_drop(x)

        # reshape x to origin shape
        x = rearrange(x, "(B N_t) S C -> B (N_t S) C", N_t=N_t)

        return x



# Modified from ``https://github.com/openai/CLIP'' and ``https://github.com/mlfoundations/open_clip''
# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
import logging
import math

import torch
import torch.nn as nn
import torch.nn.functional as F
import torchvision.transforms as T

from .tokenizers import HuggingfaceTokenizer
from .xlm_roberta import XLMRoberta
from ..kernels.attn import attention

__all__ = [
    "XLMRobertaCLIP",
    "clip_xlm_roberta_vit_h_14",
    "CLIPModel",
]


def pos_interpolate(pos, seq_len):
    if pos.size(1) == seq_len:
        return pos
    else:
        src_grid = int(math.sqrt(pos.size(1)))
        tar_grid = int(math.sqrt(seq_len))
        n = pos.size(1) - src_grid * src_grid
        return torch.cat(
            [
                pos[:, :n],
                F.interpolate(
                    pos[:, n:]
                    .float()
                    .reshape(1, src_grid, src_grid, -1)
                    .permute(0, 3, 1, 2),
                    size=(tar_grid, tar_grid),
                    mode="bicubic",
                    align_corners=False,
                )
                .flatten(2)
                .transpose(1, 2),
            ],
            dim=1,
        )


class QuickGELU(nn.Module):
    def forward(self, x):
        return x * torch.sigmoid(1.702 * x)


class LayerNorm(nn.LayerNorm):
    def forward(self, x):
        return super().forward(x.float()).type_as(x)


class SelfAttention(nn.Module):
    def __init__(
        self, dim, num_heads, causal=False, attn_dropout=0.0, proj_dropout=0.0
    ):
        assert dim % num_heads == 0
        super().__init__()
        self.dim = dim
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.causal = causal
        self.attn_dropout = attn_dropout
        self.proj_dropout = proj_dropout

        # layers
        self.to_qkv = nn.Linear(dim, dim * 3)
        self.proj = nn.Linear(dim, dim)

    def forward(self, x):
        """
        x:   [B, L, C].
        """
        b, s, c, n, d = *x.size(), self.num_heads, self.head_dim

        # compute query, key, value
        q, k, v = self.to_qkv(x).view(b, s, 3, n, d).unbind(2)

        x = attention(q, k, v)
        x = x.reshape(b, s, c)

        # output
        x = self.proj(x)
        x = F.dropout(x, self.proj_dropout, self.training)
        return x


class SwiGLU(nn.Module):
    def __init__(self, dim, mid_dim):
        super().__init__()
        self.dim = dim
        self.mid_dim = mid_dim

        # layers
        self.fc1 = nn.Linear(dim, mid_dim)
        self.fc2 = nn.Linear(dim, mid_dim)
        self.fc3 = nn.Linear(mid_dim, dim)

    def forward(self, x):
        x = F.silu(self.fc1(x)) * self.fc2(x)
        x = self.fc3(x)
        return x


class AttentionBlock(nn.Module):
    def __init__(
        self,
        dim,
        mlp_ratio,
        num_heads,
        post_norm=False,
        causal=False,
        activation="quick_gelu",
        attn_dropout=0.0,
        proj_dropout=0.0,
        norm_eps=1e-5,
    ):
        assert activation in ["quick_gelu", "gelu", "swi_glu"]
        super().__init__()
        self.dim = dim
        self.mlp_ratio = mlp_ratio
        self.num_heads = num_heads
        self.post_norm = post_norm
        self.causal = causal
        self.norm_eps = norm_eps

        # layers
        self.norm1 = LayerNorm(dim, eps=norm_eps)
        self.attn = SelfAttention(dim, num_heads, causal, attn_dropout, proj_dropout)
        self.norm2 = LayerNorm(dim, eps=norm_eps)
        if activation == "swi_glu":
            self.mlp = SwiGLU(dim, int(dim * mlp_ratio))
        else:
            self.mlp = nn.Sequential(
                nn.Linear(dim, int(dim * mlp_ratio)),
                QuickGELU() if activation == "quick_gelu" else nn.GELU(),
                nn.Linear(int(dim * mlp_ratio), dim),
                nn.Dropout(proj_dropout),
            )

    def forward(self, x):
        if self.post_norm:
            x = x + self.norm1(self.attn(x))
            x = x + self.norm2(self.mlp(x))
        else:
            x = x + self.attn(self.norm1(x))
            x = x + self.mlp(self.norm2(x))
        return x


class AttentionPool(nn.Module):
    def __init__(
        self,
        dim,
        mlp_ratio,
        num_heads,
        activation="gelu",
        proj_dropout=0.0,
        norm_eps=1e-5,
    ):
        assert dim % num_heads == 0
        super().__init__()
        self.dim = dim
        self.mlp_ratio = mlp_ratio
        self.num_heads = num_heads
        self.head_dim = dim // num_heads
        self.proj_dropout = proj_dropout
        self.norm_eps = norm_eps

        # layers
        gain = 1.0 / math.sqrt(dim)
        self.cls_embedding = nn.Parameter(gain * torch.randn(1, 1, dim))
        self.to_q = nn.Linear(dim, dim)
        self.to_kv = nn.Linear(dim, dim * 2)
        self.proj = nn.Linear(dim, dim)
        self.norm = LayerNorm(dim, eps=norm_eps)
        self.mlp = nn.Sequential(
            nn.Linear(dim, int(dim * mlp_ratio)),
            QuickGELU() if activation == "quick_gelu" else nn.GELU(),
            nn.Linear(int(dim * mlp_ratio), dim),
            nn.Dropout(proj_dropout),
        )

    def forward(self, x):
        """
        x:  [B, L, C].
        """
        b, s, c, n, d = *x.size(), self.num_heads, self.head_dim

        # compute query, key, value
        q = self.to_q(self.cls_embedding).view(1, 1, n, d).expand(b, -1, -1, -1)
        k, v = self.to_kv(x).view(b, s, 2, n, d).unbind(2)

        # compute attention
        x = flash_attn_func(q, k, v)
        x = x.reshape(b, 1, c)

        # output
        x = self.proj(x)
        x = F.dropout(x, self.proj_dropout, self.training)

        # mlp
        x = x + self.mlp(self.norm(x))
        return x[:, 0]


class VisionTransformer(nn.Module):
    def __init__(
        self,
        image_size=224,
        patch_size=16,
        dim=768,
        mlp_ratio=4,
        out_dim=512,
        num_heads=12,
        num_layers=12,
        pool_type="token",
        pre_norm=True,
        post_norm=False,
        activation="quick_gelu",
        attn_dropout=0.0,
        proj_dropout=0.0,
        embedding_dropout=0.0,
        norm_eps=1e-5,
    ):
        if image_size % patch_size != 0:
            print("[WARNING] image_size is not divisible by patch_size", flush=True)
        assert pool_type in ("token", "token_fc", "attn_pool")
        out_dim = out_dim or dim
        super().__init__()
        self.image_size = image_size
        self.patch_size = patch_size
        self.num_patches = (image_size // patch_size) ** 2
        self.dim = dim
        self.mlp_ratio = mlp_ratio
        self.out_dim = out_dim
        self.num_heads = num_heads
        self.num_layers = num_layers
        self.pool_type = pool_type
        self.post_norm = post_norm
        self.norm_eps = norm_eps

        # embeddings
        gain = 1.0 / math.sqrt(dim)
        self.patch_embedding = nn.Conv2d(
            3, dim, kernel_size=patch_size, stride=patch_size, bias=not pre_norm
        )
        if pool_type in ("token", "token_fc"):
            self.cls_embedding = nn.Parameter(gain * torch.randn(1, 1, dim))
        self.pos_embedding = nn.Parameter(
            gain
            * torch.randn(
                1,
                self.num_patches + (1 if pool_type in ("token", "token_fc") else 0),
                dim,
            )
        )
        self.dropout = nn.Dropout(embedding_dropout)

        # transformer
        self.pre_norm = LayerNorm(dim, eps=norm_eps) if pre_norm else None
        self.transformer = nn.Sequential(
            *[
                AttentionBlock(
                    dim,
                    mlp_ratio,
                    num_heads,
                    post_norm,
                    False,
                    activation,
                    attn_dropout,
                    proj_dropout,
                    norm_eps,
                )
                for _ in range(num_layers)
            ]
        )
        self.post_norm = LayerNorm(dim, eps=norm_eps)

        # head
        if pool_type == "token":
            self.head = nn.Parameter(gain * torch.randn(dim, out_dim))
        elif pool_type == "token_fc":
            self.head = nn.Linear(dim, out_dim)
        elif pool_type == "attn_pool":
            self.head = AttentionPool(
                dim, mlp_ratio, num_heads, activation, proj_dropout, norm_eps
            )

    def forward(self, x, interpolation=False, use_31_block=False):
        b = x.size(0)

        # embeddings
        x = self.patch_embedding(x).flatten(2).permute(0, 2, 1)
        if self.pool_type in ("token", "token_fc"):
            x = torch.cat([self.cls_embedding.expand(b, -1, -1), x], dim=1)
        if interpolation:
            e = pos_interpolate(self.pos_embedding, x.size(1))
        else:
            e = self.pos_embedding
        x = self.dropout(x + e)
        if self.pre_norm is not None:
            x = self.pre_norm(x)

        # transformer
        if use_31_block:
            x = self.transformer[:-1](x)
            return x
        else:
            x = self.transformer(x)
            return x


class XLMRobertaWithHead(XLMRoberta):
    def __init__(self, **kwargs):
        self.out_dim = kwargs.pop("out_dim")
        super().__init__(**kwargs)

        # head
        mid_dim = (self.dim + self.out_dim) // 2
        self.head = nn.Sequential(
            nn.Linear(self.dim, mid_dim, bias=False),
            nn.GELU(),
            nn.Linear(mid_dim, self.out_dim, bias=False),
        )

    def forward(self, ids):
        # xlm-roberta
        x = super().forward(ids)

        # average pooling
        mask = ids.ne(self.pad_id).unsqueeze(-1).to(x)
        x = (x * mask).sum(dim=1) / mask.sum(dim=1)

        # head
        x = self.head(x)
        return x


class XLMRobertaCLIP(nn.Module):
    def __init__(
        self,
        embed_dim=1024,
        image_size=224,
        patch_size=14,
        vision_dim=1280,
        vision_mlp_ratio=4,
        vision_heads=16,
        vision_layers=32,
        vision_pool="token",
        vision_pre_norm=True,
        vision_post_norm=False,
        activation="gelu",
        vocab_size=250002,
        max_text_len=514,
        type_size=1,
        pad_id=1,
        text_dim=1024,
        text_heads=16,
        text_layers=24,
        text_post_norm=True,
        text_dropout=0.1,
        attn_dropout=0.0,
        proj_dropout=0.0,
        embedding_dropout=0.0,
        norm_eps=1e-5,
    ):
        super().__init__()
        self.embed_dim = embed_dim
        self.image_size = image_size
        self.patch_size = patch_size
        self.vision_dim = vision_dim
        self.vision_mlp_ratio = vision_mlp_ratio
        self.vision_heads = vision_heads
        self.vision_layers = vision_layers
        self.vision_pre_norm = vision_pre_norm
        self.vision_post_norm = vision_post_norm
        self.activation = activation
        self.vocab_size = vocab_size
        self.max_text_len = max_text_len
        self.type_size = type_size
        self.pad_id = pad_id
        self.text_dim = text_dim
        self.text_heads = text_heads
        self.text_layers = text_layers
        self.text_post_norm = text_post_norm
        self.norm_eps = norm_eps

        # models
        self.visual = VisionTransformer(
            image_size=image_size,
            patch_size=patch_size,
            dim=vision_dim,
            mlp_ratio=vision_mlp_ratio,
            out_dim=embed_dim,
            num_heads=vision_heads,
            num_layers=vision_layers,
            pool_type=vision_pool,
            pre_norm=vision_pre_norm,
            post_norm=vision_post_norm,
            activation=activation,
            attn_dropout=attn_dropout,
            proj_dropout=proj_dropout,
            embedding_dropout=embedding_dropout,
            norm_eps=norm_eps,
        )
        self.textual = XLMRobertaWithHead(
            vocab_size=vocab_size,
            max_seq_len=max_text_len,
            type_size=type_size,
            pad_id=pad_id,
            dim=text_dim,
            out_dim=embed_dim,
            num_heads=text_heads,
            num_layers=text_layers,
            post_norm=text_post_norm,
            dropout=text_dropout,
        )
        self.log_scale = nn.Parameter(math.log(1 / 0.07) * torch.ones([]))

    def forward(self, imgs, txt_ids):
        """
        imgs:       [B, 3, H, W] of torch.float32.
        - mean:     [0.48145466, 0.4578275, 0.40821073]
        - std:      [0.26862954, 0.26130258, 0.27577711]
        txt_ids:    [B, L] of torch.long.
                    Encoded by data.CLIPTokenizer.
        """
        xi = self.visual(imgs)
        xt = self.textual(txt_ids)
        return xi, xt

    def param_groups(self):
        groups = [
            {
                "params": [
                    p
                    for n, p in self.named_parameters()
                    if "norm" in n or n.endswith("bias")
                ],
                "weight_decay": 0.0,
            },
            {
                "params": [
                    p
                    for n, p in self.named_parameters()
                    if not ("norm" in n or n.endswith("bias"))
                ]
            },
        ]
        return groups


def _clip(
    pretrained=False,
    pretrained_name=None,
    model_cls=XLMRobertaCLIP,
    return_transforms=False,
    return_tokenizer=False,
    tokenizer_padding="eos",
    dtype=torch.float32,
    device="cpu",
    **kwargs,
):
    # init a model on device
    with torch.device(device):
        model = model_cls(**kwargs)

    # set device
    model = model.to(dtype=dtype, device=device)
    output = (model,)

    # init transforms
    if return_transforms:
        # mean and std
        if "siglip" in pretrained_name.lower():
            mean, std = [0.5, 0.5, 0.5], [0.5, 0.5, 0.5]
        else:
            mean = [0.48145466, 0.4578275, 0.40821073]
            std = [0.26862954, 0.26130258, 0.27577711]

        # transforms
        transforms = T.Compose(
            [
                T.Resize(
                    (model.image_size, model.image_size),
                    interpolation=T.InterpolationMode.BICUBIC,
                ),
                T.ToTensor(),
                T.Normalize(mean=mean, std=std),
            ]
        )
        output += (transforms,)
    return output[0] if len(output) == 1 else output


def clip_xlm_roberta_vit_h_14(
    pretrained=False,
    pretrained_name="open-clip-xlm-roberta-large-vit-huge-14",
    **kwargs,
):
    cfg = dict(
        embed_dim=1024,
        image_size=224,
        patch_size=14,
        vision_dim=1280,
        vision_mlp_ratio=4,
        vision_heads=16,
        vision_layers=32,
        vision_pool="token",
        activation="gelu",
        vocab_size=250002,
        max_text_len=514,
        type_size=1,
        pad_id=1,
        text_dim=1024,
        text_heads=16,
        text_layers=24,
        text_post_norm=True,
        text_dropout=0.1,
        attn_dropout=0.0,
        proj_dropout=0.0,
        embedding_dropout=0.0,
    )
    cfg.update(**kwargs)
    return _clip(pretrained, pretrained_name, XLMRobertaCLIP, **cfg)


class CLIPModel:
    def __init__(self, dtype, device, checkpoint_path, tokenizer_path):
        self.dtype = dtype
        self.device = device
        self.checkpoint_path = checkpoint_path
        self.tokenizer_path = tokenizer_path

        # init model
        self.model, self.transforms = clip_xlm_roberta_vit_h_14(
            pretrained=False,
            return_transforms=True,
            return_tokenizer=False,
            dtype=dtype,
            device=device,
        )
        self.model = self.model.eval().requires_grad_(False)
        logging.info(f"loading {checkpoint_path}")
        self.model.load_state_dict(torch.load(checkpoint_path, map_location="cpu"))

        # init tokenizer
        self.tokenizer = HuggingfaceTokenizer(
            name=tokenizer_path, seq_len=self.model.max_text_len - 2, clean="whitespace"
        )

    def visual(self, videos):
        # preprocess
        size = (self.model.image_size,) * 2
        videos = torch.cat(
            [
                F.interpolate(
                    u.transpose(0, 1), size=size, mode="bicubic", align_corners=False
                )
                for u in videos
            ]
        )
        videos = self.transforms.transforms[-1](videos.mul_(0.5).add_(0.5))

        # forward
        with torch.amp.autocast("cuda", dtype=self.dtype):
            out = self.model.visual(videos, use_31_block=True)
            return out



# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
import torch
from easydict import EasyDict

# ------------------------ Wan MultiTalk AI2V 14B ------------------------#
multitalk_14B = EasyDict(__name__="Config: Wan MultiTalk AI2V 14B")

# t5
multitalk_14B.t5_model = "umt5_xxl"
multitalk_14B.t5_dtype = torch.bfloat16
multitalk_14B.text_len = 512

# transformer
multitalk_14B.param_dtype = torch.bfloat16

# inference
multitalk_14B.num_train_timesteps = 1000
multitalk_14B.sample_fps = 16
multitalk_14B.sample_neg_prompt = "bright tones, overexposed, static, blurred details, subtitles, style, works, paintings, images, static, overall gray, worst quality, low quality, JPEG compression residue, ugly, incomplete, extra fingers, poorly drawn hands, poorly drawn faces, deformed, disfigured, misshapen limbs, fused fingers, still picture, messy background, three legs, many people in the background, walking backwards"

multitalk_14B.t5_checkpoint = "models_t5_umt5-xxl-enc-bf16.pth"
multitalk_14B.t5_tokenizer = "google/umt5-xxl"

# clip
multitalk_14B.clip_model = "clip_xlm_roberta_vit_h_14"
multitalk_14B.clip_dtype = torch.float16
multitalk_14B.clip_checkpoint = (
    "models_clip_open-clip-xlm-roberta-large-vit-huge-14.pth"
)
multitalk_14B.clip_tokenizer = "xlm-roberta-large"

# vae
multitalk_14B.vae_checkpoint = "Wan2.1_VAE.pth"
multitalk_14B.vae_stride = (4, 8, 8)

# transformer
multitalk_14B.patch_size = (1, 2, 2)
multitalk_14B.dim = 5120
multitalk_14B.ffn_dim = 13824
multitalk_14B.freq_dim = 256
multitalk_14B.num_heads = 40
multitalk_14B.num_layers = 40
multitalk_14B.window_size = (-1, -1)
multitalk_14B.qk_norm = True
multitalk_14B.cross_attn_norm = True
multitalk_14B.eps = 1e-6



# Copyright 2024-2025 The Alibaba Wan Team Authors. All rights reserved.
import math
import os
import time
import yaml
from typing import Literal
from collections import deque
from pathlib import Path

import numpy as np
import torch
import torch.nn as nn
from einops import rearrange
from transformers import Wav2Vec2FeatureExtractor
from loguru import logger

from osc_data.video import Video
from osc_data.image import Image
from osc_data.audio import Audio

from .models import WanVAE, CLIPModel, T5EncoderModel, Wav2Vec2Model
from .models.dit import WanModel, WanLayerNorm, WanRMSNorm
from .utils import (
    match_and_blend_colors_torch,
    resize_and_centercrop,
    loudness_norm,
)
from .quantize import quantize_model_a8w8_int8_gemlite
from .gemlite.core import GemLiteLinear
from .vram_management import (
    enable_vram_management,
    AutoWrappedLinear,
    AutoWrappedModule,
)
from .configs import multitalk_14B

# compile models to speedup inference
COMPILE_MODEL = True
COMPILE_VAE = True
# use parallel vae to speedup decode/encode
USE_PARALLEL_VAE = True


def to_param_dtype_fp32only(model, param_dtype):
    for module in model.modules():
        for name, param in module.named_parameters(recurse=False):
            if param.dtype == torch.float32 and param.__class__.__name__ not in [
                "WeightQBytesTensor"
            ]:
                param.data = param.data.to(param_dtype)
        for name, buf in module.named_buffers(recurse=False):
            if buf.dtype == torch.float32 and buf.__class__.__name__ not in [
                "WeightQBytesTensor"
            ]:
                module._buffers[name] = buf.to(param_dtype)


def timestep_transform(
    t,
    shift=5.0,
    num_timesteps=1000,
):
    t = t / num_timesteps
    # shift the timestep based on ratio
    new_t = shift * t / (1 + (shift - 1) * t)
    new_t = new_t * num_timesteps
    return new_t


class FlashTalkPipeline:
    def __init__(
        self,
        checkpoint_dir,
        wav2vec_dir,
        device="cuda",
        num_timesteps=1000,
        use_timestep_transform=True,
        num_persistent_param_in_dit=15_000_000_000,
    ):
        """FlashTalkPipeline for RTX 4090 GPU.
        Args:
            checkpoint_dir (`str`):
                Path to directory containing model checkpoints
            wav2vec_dir (`str`):
                Path to directory containing wav2vec checkpoints
            num_timesteps (`int`, *optional*, defaults to 1000):
                Number of timesteps.
            use_timestep_transform (`bool`, *optional*, defaults to True):
                Enable timestep transform.
            num_persistent_param_in_dit (`int`, *optional*, defaults to 15_000_000_000):
                Number of persistent parameters in DIT model.
        """
        self.device = device
        config = multitalk_14B
        self.config = config

        with open(Path(__file__).parent / "configs" / "infer_params.yaml", "r") as f:
            self.infer_params = yaml.safe_load(f)
        self.rank = 0
        self.use_usp = False
        self.param_dtype = config.param_dtype
        self.cpu_offload = True

        self.text_encoder = T5EncoderModel(
            text_len=config.text_len,
            dtype=config.t5_dtype,
            device="cpu" if self.cpu_offload else self.device,
            checkpoint_path=os.path.join(checkpoint_dir, config.t5_checkpoint),
            tokenizer_path=os.path.join(checkpoint_dir, config.t5_tokenizer),
        )

        self.vae_stride = config.vae_stride
        self.patch_size = config.patch_size

        self.vae = WanVAE(
            vae_path=os.path.join(checkpoint_dir, config.vae_checkpoint),
            dtype=self.param_dtype,
            device="cpu" if self.cpu_offload else self.device,
            parallel=False,
        )

        self.clip = CLIPModel(
            dtype=config.clip_dtype,
            device="cpu" if self.cpu_offload else self.device,
            checkpoint_path=os.path.join(checkpoint_dir, config.clip_checkpoint),
            tokenizer_path=os.path.join(checkpoint_dir, config.clip_tokenizer),
        )

        logger.info(f"Creating WanModel from {checkpoint_dir}")

        self.model = WanModel.from_pretrained(
            checkpoint_dir,
            device_map="cpu" if self.cpu_offload else self.device,
            torch_dtype=self.param_dtype,
        )
        self.model.eval().requires_grad_(False)
        quantize_model_a8w8_int8_gemlite(self.model, device="cuda")
        self.model.cpu()
        torch.cuda.empty_cache()
        logger.info(
            f"Enable low vram mode with num_persistent_param_in_dit: {num_persistent_param_in_dit}"
        )
        self.vram_management = False
        self.enable_vram_management(
            num_persistent_param_in_dit=num_persistent_param_in_dit
        )

        self.sample_neg_prompt = config.sample_neg_prompt
        self.num_timesteps = num_timesteps
        self.use_timestep_transform = use_timestep_transform

        self.audio_encoder = Wav2Vec2Model.from_pretrained(
            wav2vec_dir, local_files_only=True
        ).to("cpu")
        self.audio_encoder.feature_extractor._freeze_parameters()
        self.wav2vec_feature_extractor = Wav2Vec2FeatureExtractor.from_pretrained(
            wav2vec_dir, local_files_only=True
        )

        self.model_names = ["model"]

    def enable_vram_management(self, num_persistent_param_in_dit=0):
        dtype = next(iter(self.model.parameters())).dtype
        logger.info(f"Enable vram management with dtype: {dtype}")
        enable_vram_management(
            self.model,
            module_map={
                GemLiteLinear: AutoWrappedModule,
                torch.nn.Linear: AutoWrappedLinear,
                torch.nn.Conv3d: AutoWrappedModule,
                torch.nn.LayerNorm: AutoWrappedModule,
                WanLayerNorm: AutoWrappedModule,
                WanRMSNorm: AutoWrappedModule,
            },
            module_config=dict(
                offload_dtype=dtype,
                offload_device="cpu",
                onload_dtype=dtype,
                onload_device=self.device,
                computation_dtype=self.param_dtype,
                computation_device=self.device,
            ),
            max_num_param=num_persistent_param_in_dit,
            overflow_module_config=dict(
                offload_dtype=dtype,
                offload_device="cpu",
                onload_dtype=dtype,
                onload_device="cpu",
                computation_dtype=self.param_dtype,
                computation_device=self.device,
            ),
        )
        self.enable_cpu_offload()
        self.vram_management = True

    def enable_cpu_offload(self):
        self.cpu_offload = True

    def onload_dit_model(self):
        start_time = time.time()
        if not self.cpu_offload:
            return
        for model_name in self.model_names:
            model = getattr(self, model_name)
            if not isinstance(model, nn.Module):
                model = model.model
            if model is not None:
                if (
                    hasattr(model, "vram_management_enabled")
                    and model.vram_management_enabled
                ):
                    for module in model.modules():
                        if hasattr(module, "onload"):
                            module.onload()
                else:
                    model.to(self.device)
        end_time = time.time()
        logger.info(f"Onload dit model time: {end_time - start_time} seconds")

    def offload_dit_model(self):
        start_time = time.time()
        if not self.cpu_offload:
            return
        for model_name in self.model_names:
            model = getattr(self, model_name)
            if not isinstance(model, nn.Module):
                model = model.model
            if model is not None:
                if (
                    hasattr(model, "vram_management_enabled")
                    and model.vram_management_enabled
                ):
                    for module in model.modules():
                        if hasattr(module, "offload"):
                            module.offload()
                else:
                    model.to(self.device)
        torch.cuda.empty_cache()
        end_time = time.time()
        logger.info(f"Offload dit model time: {end_time - start_time} seconds")

    @torch.no_grad()
    def prepare_params(
        self,
        input_prompt: str,
        cond_image: Image,
        target_size: tuple[int, int] = (768, 448),
        frame_num: int = 33,
        motion_frames_num: int = 5,
        sampling_steps: int = 4,
        seed: int = 9999,
        shift: int = 5,
        color_correction_strength: float = 1.0,
    ):

        context = self.text_encoder(input_prompt, self.device)[0].to(self.device)

        self.frame_num = frame_num
        self.motion_frames_num = motion_frames_num
        self.target_h, self.target_w = target_size
        cond_image_tensor = torch.from_numpy(cond_image.load().to_rgb().data).permute(
            2, 0, 1
        )  # C, H, W
        # if isinstance(cond_image, str):
        #     cond_image = Image.open(cond_image).convert("RGB")
        cond_image_tensor = resize_and_centercrop(
            cond_image_tensor, (self.target_h, self.target_w)
        ).to(dtype=self.param_dtype, device=self.device)
        cond_image_tensor = (cond_image_tensor / 255 - 0.5) * 2
        self.cond_image_tensor = cond_image_tensor

        self.color_correction_strength = color_correction_strength
        self.original_color_reference = None
        if self.color_correction_strength > 0.0:
            self.original_color_reference = cond_image_tensor.clone()

        if self.cpu_offload:
            self.clip.model.to(self.device)
        clip_context = self.clip.visual(cond_image_tensor[:, :, -1:, :, :]).to(
            self.param_dtype
        )
        if self.cpu_offload:
            self.clip.model.cpu()
            torch.cuda.empty_cache()
        video_frames = torch.zeros(
            1,
            cond_image_tensor.shape[1],
            frame_num - cond_image_tensor.shape[2],
            self.target_h,
            self.target_w,
        ).to(dtype=self.param_dtype, device=self.device)
        padding_frames_pixels_values = torch.concat(
            [cond_image_tensor, video_frames], dim=2
        )
        if self.cpu_offload:
            self.vae.model.to(self.device)
            self.vae.scale[0] = self.vae.scale[0].to(self.device)
            self.vae.scale[1] = self.vae.scale[1].to(self.device)
        y = self.vae.encode(padding_frames_pixels_values)
        common_y = y.unsqueeze(0).to(self.param_dtype)

        # get mask
        self.lat_h, self.lat_w = (
            self.target_h // self.vae_stride[1],
            self.target_w // self.vae_stride[2],
        )
        msk = torch.ones(1, frame_num, self.lat_h, self.lat_w, device=self.device)
        msk[:, 1:] = 0
        msk = torch.concat(
            [torch.repeat_interleave(msk[:, 0:1], repeats=4, dim=1), msk[:, 1:]], dim=1
        )
        msk = msk.view(1, msk.shape[1] // 4, 4, self.lat_h, self.lat_w)
        msk = msk.transpose(1, 2).to(self.param_dtype)

        y = torch.concat([msk, common_y], dim=1)

        max_seq_len = (
            ((frame_num - 1) // self.vae_stride[0] + 1)
            * self.lat_h
            * self.lat_w
            // (self.patch_size[1] * self.patch_size[2])
        )
        max_seq_len = int(math.ceil(max_seq_len / 1)) * 1

        self.generator = torch.Generator(device=self.device).manual_seed(seed)

        # prepare timesteps
        if sampling_steps == 2:
            timesteps = [1000, 500]
        elif sampling_steps == 4:
            timesteps = [1000, 750, 500, 250]
        else:
            timesteps = list(
                np.linspace(self.num_timesteps, 1, sampling_steps, dtype=np.float32)
            )

        timesteps.append(0.0)
        timesteps = [torch.tensor([t], device=self.device) for t in timesteps]
        if self.use_timestep_transform:
            timesteps = [
                timestep_transform(t, shift=shift, num_timesteps=self.num_timesteps)
                for t in timesteps
            ]
        self.timesteps = timesteps

        self.arg_c = {
            "context": [context],
            "clip_fea": clip_context,
            "seq_len": max_seq_len,
            "y": y,
            "ref_target_masks": None,
        }

        self.latent_motion_frames = self.vae.encode(self.cond_image_tensor)

        if self.cpu_offload:
            self.vae.model.cpu()
            torch.cuda.empty_cache()

        return

    @torch.no_grad()
    def preprocess_audio(self, speech_array, sr: int = 16000, fps: int = 25):
        video_length = len(speech_array) * fps / sr

        # wav2vec_feature_extractor
        audio_feature = np.squeeze(
            self.wav2vec_feature_extractor(speech_array, sampling_rate=sr).input_values
        )
        audio_feature = torch.from_numpy(audio_feature).float().to(device=self.device)
        audio_feature = audio_feature.unsqueeze(0)

        # audio encoder
        if self.cpu_offload:
            self.audio_encoder.to(self.device)
        with torch.no_grad():
            embeddings = self.audio_encoder(
                audio_feature, seq_len=int(video_length), output_hidden_states=True
            )
        if self.cpu_offload:
            self.audio_encoder.cpu()
            torch.cuda.empty_cache()

        if len(embeddings) == 0:
            logger.error("Fail to extract audio embedding")
            return None

        audio_emb = torch.stack(embeddings.hidden_states[1:], dim=1).squeeze(0)
        audio_emb = rearrange(audio_emb, "b s d -> s b d")
        return audio_emb

    def get_audio_embedding(self, audio_array, audio_start_idx=-1, audio_end_idx=-1):
        audio_array = loudness_norm(audio_array, self.infer_params["sample_rate"])
        audio_embedding = self.preprocess_audio(
            audio_array,
            sr=self.infer_params["sample_rate"],
            fps=self.infer_params["tgt_fps"],
        )

        if audio_start_idx == -1 or audio_end_idx == -1:
            audio_start_idx = 0
            audio_end_idx = audio_embedding.shape[0]

        indices = (torch.arange(2 * 2 + 1) - 2) * 1

        center_indices = torch.arange(audio_start_idx, audio_end_idx, 1).unsqueeze(
            1
        ) + indices.unsqueeze(0)
        center_indices = torch.clamp(center_indices, min=0, max=audio_end_idx - 1)

        audio_embedding = audio_embedding[center_indices][None, ...].contiguous()
        return audio_embedding

    @torch.no_grad()
    def generate_chunk(self, audio_embedding):
        """
        Generate a chunk of video from the audio embedding.
        Args:
            audio_embedding: The audio embedding.
        Returns:
            The generated video
        """
        # evaluation mode
        with torch.no_grad():
            self.arg_c.update(
                {
                    "audio": audio_embedding,
                }
            )

            # sample videos
            latent = torch.randn(
                16,
                (self.frame_num - 1) // 4 + 1,
                self.lat_h,
                self.lat_w,
                dtype=self.param_dtype,
                device=self.device,
                generator=self.generator,
            )

            latent[:, : self.latent_motion_frames.shape[1]] = self.latent_motion_frames

            for i in range(len(self.timesteps) - 1):
                timestep = self.timesteps[i]
                latent_model_input = [latent]
                # onload dit model
                if self.vram_management:
                    self.onload_dit_model()

                # inference without CFG
                start_time = time.perf_counter()
                noise_pred_cond = self.model(
                    latent_model_input, t=timestep, **self.arg_c
                )[0]
                torch.cuda.synchronize()
                end_time = time.perf_counter()
                logger.info(f"model denoise per step: {end_time - start_time:.2f}s")

                noise_pred = -noise_pred_cond

                # update latent
                t_i = self.timesteps[i][:, None, None, None] / self.num_timesteps
                t_i_1 = self.timesteps[i + 1][:, None, None, None] / self.num_timesteps
                x_0 = latent + noise_pred * t_i

                latent = (1 - t_i_1) * x_0 + t_i_1 * torch.randn(
                    x_0.size(),
                    dtype=x_0.dtype,
                    device=self.device,
                    generator=self.generator,
                )

                latent[:, : self.latent_motion_frames.shape[1]] = (
                    self.latent_motion_frames
                )

            # self.offload_dit_model()

            if self.cpu_offload:
                self.vae.model.to(self.device)
                self.vae.scale[0] = self.vae.scale[0].to(self.device)
                self.vae.scale[1] = self.vae.scale[1].to(self.device)

            torch.cuda.synchronize()
            start_decode_time = time.time()
            videos = self.vae.decode(latent.to(self.param_dtype))
            torch.cuda.synchronize()
            end_decode_time = time.time()
            logger.info(
                f"decode video frames: {end_decode_time - start_decode_time:.2f}s"
            )

        torch.cuda.synchronize()
        start_color_correction_time = time.time()
        if self.color_correction_strength > 0.0:
            videos = match_and_blend_colors_torch(
                videos, self.original_color_reference, self.color_correction_strength
            )

        cond_frame = videos[:, :, -self.motion_frames_num :].to(self.device)
        torch.cuda.synchronize()
        end_color_correction_time = time.time()
        logger.info(
            f"color correction: {end_color_correction_time - start_color_correction_time:.2f}s"
        )

        torch.cuda.synchronize()
        start_encode_time = time.time()
        self.latent_motion_frames = self.vae.encode(cond_frame)
        torch.cuda.synchronize()
        end_encode_time = time.time()
        logger.info(f"encode motion frames: {end_encode_time - start_encode_time:.2f}s")

        if self.cpu_offload:
            self.vae.model.cpu()
            torch.cuda.empty_cache()

        gen_video_samples = videos  # [:, :, self.motion_frames_num:]
        gen_video_samples = gen_video_samples[0].to(torch.float32)
        gen_video_samples = (
            ((gen_video_samples + 1) / 2).permute(1, 2, 3, 0).clip(0, 1) * 255
        ).contiguous()
        return gen_video_samples

    def generate(
        self,
        input_prompt: str,
        audio: Audio,
        image: Image,
        audio_encode_mode: Literal["stream", "once"] = "once",
    ) -> Video:
        """
        Generate video from the audio and image prompt.

        This method processes audio input along with a conditioning image to generate
        a talking head video. The audio is processed in chunks and each chunk generates
        corresponding video frames.

        Args:
            input_prompt: The input text prompt describing the desired video generation.
            image: The conditioning image used as visual reference.
            audio: The audio file (wav format) that drives the lip movements.
            audio_encode_mode: Strategy for encoding audio, either "stream" or "once".
                - "stream": Process audio in streaming chunks (default), memory efficient.
                - "once": Encode entire audio at once then split into chunks.
            video_save_path: Path to save the generated video.
            merge_video_audio: Whether to merge the generated video and the original audio.
            force_9_16: Whether to force the video to be 9:16.

        Returns:
            List[torch.Tensor]: A list of video frame tensors, where each tensor has:
                - Shape: (num_frames, height, width, 3)
                - Dtype: torch.float32
                - Value range: [0, 255]
                - Device: CPU
                - Channel order: RGB (channels last)

                For example, with 768x448 resolution, each tensor shape is:
                - First chunk: (33, 768, 448, 3)
                - Subsequent chunks: (28, 768, 448, 3) or (23, 768, 448, 3) for last chunk
        """
        generate_start_time = time.perf_counter()
        logger.info("Start to generate video...")
        # prepare data
        sample_rate = self.infer_params["sample_rate"]
        tgt_fps = self.infer_params["tgt_fps"]
        cached_audio_duration = self.infer_params["cached_audio_duration"]
        frame_num = self.infer_params["frame_num"]
        motion_frames_num = self.infer_params["motion_frames_num"]
        slice_len = frame_num - motion_frames_num
        self.prepare_params(input_prompt=input_prompt, cond_image=image)

        human_speech_array_all = audio.load(
            sample_rate=self.infer_params["sample_rate"], mono=True
        ).data
        human_speech_array_slice_len = slice_len * sample_rate // tgt_fps
        human_speech_array_frame_num = frame_num * sample_rate // tgt_fps
        data_prepare_time = time.perf_counter()
        logger.info(
            f"Data preparation time: {data_prepare_time - generate_start_time:.2f}s"
        )

        generated_list = []
        if audio_encode_mode == "once":
            # pad audio with silence to avoid truncating the last chunk
            remainder = (
                len(human_speech_array_all) - human_speech_array_frame_num
            ) % human_speech_array_slice_len
            if remainder > 0:
                pad_length = human_speech_array_slice_len - remainder
                human_speech_array_all = np.concatenate(
                    [
                        human_speech_array_all,
                        np.zeros(pad_length, dtype=human_speech_array_all.dtype),
                    ]
                )

            # encode audio together
            audio_embedding_all = self.get_audio_embedding(human_speech_array_all)

            # split audio embedding into chunks: 33, 28, 28, 28, ...
            audio_embedding_len = audio_embedding_all.shape[1]
            chunk_count = (audio_embedding_len - frame_num + slice_len) // slice_len
            audio_embedding_chunks_list = [
                audio_embedding_all[
                    :, i * slice_len : i * slice_len + frame_num
                ].contiguous()
                for i in range(chunk_count)
            ]

            for chunk_idx, audio_embedding_chunk in enumerate(
                audio_embedding_chunks_list
            ):
                torch.cuda.synchronize()
                start_time = time.time()

                # inference
                video = self.generate_chunk(audio_embedding_chunk)

                if chunk_idx != 0:
                    video = video[motion_frames_num:]

                torch.cuda.synchronize()
                end_time = time.time()
                logger.info(
                    f"Generate video chunk-{chunk_idx} done, cost time: {(end_time - start_time):.2f}s"
                )

                generated_list.append(video.cpu())
                torch.cuda.empty_cache()

        elif audio_encode_mode == "stream":
            cached_audio_length_sum = sample_rate * cached_audio_duration
            audio_end_idx = cached_audio_duration * tgt_fps
            audio_start_idx = audio_end_idx - frame_num

            audio_dq = deque(
                [0.0] * cached_audio_length_sum, maxlen=cached_audio_length_sum
            )

            # pad audio with silence to avoid truncating the last chunk
            remainder = len(human_speech_array_all) % human_speech_array_slice_len
            if remainder > 0:
                pad_length = human_speech_array_slice_len - remainder
                human_speech_array_all = np.concatenate(
                    [
                        human_speech_array_all,
                        np.zeros(pad_length, dtype=human_speech_array_all.dtype),
                    ]
                )

            # split audio embedding into chunks: 28, 28, 28, 28, ...
            human_speech_array_slices = human_speech_array_all.reshape(
                -1, human_speech_array_slice_len
            )

            for chunk_idx, human_speech_array in enumerate(human_speech_array_slices):
                # streaming encode audio chunks
                audio_dq.extend(human_speech_array.tolist())
                audio_array = np.array(audio_dq)
                audio_embedding = self.get_audio_embedding(
                    audio_array, audio_start_idx, audio_end_idx
                )

                torch.cuda.synchronize()
                start_time = time.time()

                # inference
                video = self.generate_chunk(audio_embedding)
                video = video[motion_frames_num:]

                torch.cuda.synchronize()
                end_time = time.time()
                logger.info(
                    f"Generate video chunk-{chunk_idx} done, cost time: {(end_time - start_time):.2f}s"
                )

                generated_list.append(video.cpu())
                torch.cuda.empty_cache()

        # offload dit model
        if self.vram_management:
            self.offload_dit_model()
        video_array = torch.cat(generated_list, dim=0).numpy().astype(np.uint8)
        video = Video(data=video_array, prompt=input_prompt, fps=tgt_fps)
        generate_end_time = time.perf_counter()
        logger.info(
            f"FlashTalk Pipeline Generate time: {generate_end_time - generate_start_time:.2f}s"
        )
        return video
